from rdkit import Chem
from rdkit.Chem import BRICS as Brics
from rdkit.Chem import Crippen as crip
from rdkit.Chem import Lipinski as lip
from rdkit.Chem import rdMolDescriptors as desc
from rdkit.Chem import rdmolops
from typing import List
import pandas as pd
import os
import copy

alerts = pd.read_csv(os.path.join(os.path.dirname(__file__), 'data/alert_collection.csv'))
pains = alerts.loc[alerts['rule_set_name'] == 'PAINS', 'smarts']
glaxo = alerts.loc[alerts['rule_set_name'] == 'Glaxo', 'smarts']
always_filter = {'PAINS': pains, 'glaxo': glaxo}


def get_atom_props(mol):
    """
    this function takes a mol and calculates the number of carbons, charges, and the size of the largest ring
    """

    num_carbons = 0
    num_charges = 0
    max_ring_size = 0

    # loop over all the atoms
    for at in mol.GetAtoms():
        if at.GetSymbol() == "C":
            num_carbons += 1
        if at.GetFormalCharge() != 0:
            num_charges += abs(at.GetFormalCharge())

        # I suspect that there will never be a ring larger than 100, but this is technically an arbitrary cutoff
        # this really could be optimized, but it's not rate-limiting enough to bother right now
        for size in range(3, 101):
            if at.IsInRingSize(size):
                # which is larger, lazy method
                max_ring_size = max((size, max_ring_size))

    return num_carbons, num_charges, max_ring_size


def get_filter_values(mol):
    """
    calculate the values, for a given molecule, that are used to filter
    return as a dictionary
    """

    assert isinstance(mol, Chem.Mol)

    values = {}
    values["MW"] = desc.CalcExactMolWt(mol)
    values["logP"] = crip.MolLogP(mol)
    values["HBA"] = lip.NumHAcceptors(mol)
    values["HBD"] = lip.NumHDonors(mol)
    values["tPSA"] = desc.CalcTPSA(mol)
    values["rot_bonds"] = lip.NumRotatableBonds(mol)
    values["rigid_bonds"] = mol.GetNumBonds() - values["rot_bonds"]  # assume mutual exclusion
    values["num_rings"] = lip.RingCount(mol)
    values["num_hetero_atoms"] = lip.NumHeteroatoms(mol)
    values["charge"] = rdmolops.GetFormalCharge(mol)  # trusting this charge calculation method
    values["num_carbons"], values["num_charges"], values["max_ring_size"] = get_atom_props(mol)
    try:
        values["hc_ratio"] = float(values["num_hetero_atoms"]) / float(values["num_carbons"])
    except ZeroDivisionError:
        values["hc_ratio"] = 100000000  # if there are zero carbons
    values["fc"] = len(list(Brics.FindBRICSBonds(mol)))  # how many BRICS bonds, related to complexity
    values["is_good"] = True  # default to true, but not yet observed
    atoms = [atom.GetSymbol() for atom in mol.GetAtoms()]  # get all the atoms, and make the list unique (only types)
    atoms = set(atoms)
    atoms = list(atoms)
    values["atoms"] = atoms
    values["num_chiral_centers"] = len(Chem.FindMolChiralCenters(mol, includeUnassigned=True))
    values["rejections"] = []  # empty list to store the reasons for rejection

    return values


def apply_filter(final_limits, children, must_have_patterns: List[str] = None, must_not_have_patterns: List[str] = None):

    """
    Apply the filters to a set of molecules.
    :param final_limits: The drug-like parameters to use as bounds for the filters.
    :param children: The Chem.Mol objects to actually filter.
    :param must_have_patterns: An optional list of SMARTS patterns to ensure at least one of which must be present.
    :param must_not_have_patterns: An optional list of SMARTS patterns where none can be present.
    :return: A dictionary with top-level keys being SMILES strings, and next level being filter criteria.
    """

    assert isinstance(children, list)
    assert isinstance(children[0], Chem.Mol)
    assert isinstance(final_limits, dict) or (final_limits is None)
    if not must_not_have_patterns:
        must_not_have_patterns = []

    # compute the values to check for all the children
    child_values = {}
    for child in children:
        smile = Chem.MolToSmiles(child, isomericSmiles=True)
        # avoid filtering the same smile multiple times
        if smile in child_values:
            pass
        else:
            child_values[smile] = get_filter_values(child)
    if final_limits is None:
        return child_values

    # actually filter the children here
    # record each reason why a child violates the filter rules
    for smile in child_values:
        # ranges
        if child_values[smile]["MW"] <= final_limits["MW"][0]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("MW")
        if child_values[smile]["MW"] >= final_limits["MW"][1]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("MW")
        if child_values[smile]["num_carbons"] <= final_limits["num_carbons"][0]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_carbons")
        if child_values[smile]["num_carbons"] >= final_limits["num_carbons"][1]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_carbons")
        if child_values[smile]["num_hetero_atoms"] <= final_limits["num_hetero_atoms"][0]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_hetero_atoms")
        if child_values[smile]["num_hetero_atoms"] >= final_limits["num_hetero_atoms"][1]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_hetero_atoms")
        if child_values[smile]["hc_ratio"] <= final_limits["hc_ratio"][0]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("hc_ratio")
        if child_values[smile]["hc_ratio"] >= final_limits["hc_ratio"][1]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("hc_ratio")
        if child_values[smile]["charge"] <= final_limits["charge"][0]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("charge")
        if child_values[smile]["charge"] >= final_limits["charge"][1]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("charge")
        if child_values[smile]["logP"] <= final_limits["logP"][0]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("logP")
        if child_values[smile]["logP"] >= final_limits["logP"][1]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("logP")

        # upper limits
        if child_values[smile]["HBA"] >= final_limits["HBA"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("HBA")
        if child_values[smile]["HBD"] >= final_limits["HBD"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("HBD")
        if child_values[smile]["tPSA"] >= final_limits["tPSA"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("tPSA")
        if child_values[smile]["rot_bonds"] >= final_limits["rot_bonds"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("rot_bonds")
        if child_values[smile]["rigid_bonds"] >= final_limits["rigid_bonds"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("rigid_bonds")
        if child_values[smile]["num_rings"] >= final_limits["num_rings"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_rings")
        if child_values[smile]["max_ring_size"] >= final_limits["max_ring_size"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("max_ring_size")
        if child_values[smile]["num_charges"] >= final_limits["num_charges"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_charges")
        if child_values[smile]["num_chiral_centers"] >= final_limits["num_chiral_centers"]:
            child_values[smile]["is_good"] = False
            child_values[smile]["rejections"].append("num_chiral_centers")

        # exclude smarts
        smarts_dict = copy.deepcopy(always_filter)
        smarts_dict.update({'user_excluded': must_not_have_patterns})
        for name in smarts_dict.keys():
            if filter_smarts(smile=smile, patterns=smarts_dict[name]):
                child_values[smile]["is_good"] = False
                child_values[smile]["rejections"].append(name)

        # require smarts
        if must_have_patterns:
            if not filter_smarts(smile=smile, patterns=must_have_patterns):
                child_values[smile]["is_good"] = False
                child_values[smile]["rejections"].append("missing_pattern")

    return child_values


def filter_smarts(*, smile: str, patterns: List[str]):

    """
    Check a molecule for the presence of any of a set of SMARTS patterns.
    :param smile: SMILES string of the molecule to be checked
    :param patterns: list of SMARTS patterns to look for
    :return: True if molecule DOES have AT LEAST ONE of the patterns
    """
    # get mols of the search patterns
    filter_mols = [Chem.MolFromSmarts(x, mergeHs=True) for x in patterns]
    # if there is a match, stop searching and return True
    test_mol = Chem.MolFromSmiles(smile)
    for patt in filter_mols:
        if test_mol.HasSubstructMatch(patt):
            return True
    # else, no match
    return False



"""
Modified from code by Emilie S. Henault and Jan H. Jensen 2019
"""
import random
from .jensen_selfies_crossover import mol2string, string2mol
from selfies import get_semantic_robust_alphabet
from rdkit import Chem, rdBase
rdBase.DisableLog('rdApp.error')
symbols = [symbol.strip("[]") for symbol in get_semantic_robust_alphabet()]


def mutate(mol):
    child = mol2string(mol)
    for _ in range(50):
        mutated_gene = random.randint(0, len(child) - 1)
        random_symbol_number = random.randint(0, len(symbols)-1)
        new_child = list(child)
        new_child[mutated_gene] = symbols[random_symbol_number]
        new_child_mol = string2mol(new_child)
        return new_child_mol

    return mol



import os

import deriver.fragment as fragment
from .brics_mating_rules import brics_mating_rules
from .child_filter import apply_filter
from .fragment_index import frag_index
from .lib_read import lib_read
from peewee import SqliteDatabase
from rdkit.Chem.BRICS import reverseReactions
from rdkit import Chem
import random
import logging as logger


def react(user_frag, mate_mols):
    """
    a reactor for brics fragments
    :param user_frag:
    :param mate_mols:
    :return:
    """

    for mate_frag in mate_mols:

        # list of reactions in the Brics system, try them all in a random order (to prevent systematic bias)
        random.shuffle(reverseReactions, random=random.random)
        for rxn in reverseReactions:
            # run the appropriate reactions
            ps = rxn.RunReactants((user_frag, mate_frag))
            # give up the reacted fragment
            if len(ps) > 0:
                # logger.debug(f"Mated {Chem.MolToSmiles(user_frag)} and "
                #              f"{Chem.MolToSmiles(mate_frag)}")
                for p in ps:
                    yield p[0]
            ps = rxn.RunReactants((mate_frag, user_frag))
            # give up the reacted fragment
            if len(ps) > 0:
                # logger.debug(f"Mated {Chem.MolToSmiles(user_frag)} and "
                #              f"{Chem.MolToSmiles(mate_frag)}")
                for p in ps:
                    yield p[0]

        # logger.error(f"No reactions took place between {Chem.MolToSmiles(user_frag)} and "
        #              f"{Chem.MolToSmiles(mate_frag)}")


def mate(db,
         user_smile,
         user_num_pseudo_atoms,
         missing_piece_fc,
         missing_piece_len,
         permissivity=1.0,
         max_num_children=25,
         final_limits=None,
         must_have_patterns=None,
         must_not_have_patterns=None
         ):
    """
    mate()

    'main' function for this module
    generates children using a library of processed fragments, and a specific user fragment
        db = location of the fragment library database
        user_frag = Fragment object from lib_read.py
        missing_piece_fc = the estimated FC of the pieces removed from the user parent, to create this user frag
        missing_piece_len = see above but for length
        permissivity = scaling factor that controls how deviant molecules are allowed to be:
                        1~parity to parent, 0~smallest possible additions, >1~allow adding larger/more complex fragments
        max_num_children = the number of children to return for this fragment
        max_pseudoatoms = not user facing, use caution. The starting value for how many pseudoatoms a mate frag is allowed

    :param db:
    :param user_smile:
    :param user_num_pseudo_atoms:
    :param missing_piece_fc:
    :param missing_piece_len:
    :param permissivity:
    :param max_num_children:
    :param max_pseudoatoms:
    :return:

    """

    # now get the list of pseudoatoms in the user fragment
    _, _, p_atoms = fragment.get_pseudo_atoms(user_smile)
    user_frag_mol = Chem.MolFromSmiles(user_smile)
    if user_frag_mol is None:
        logger.error(f"Could not process this fragment into a Mol object, skipping: {user_smile}")
        return [], {}

    # the final list of all the children
    finished_children = list()

    # a counter
    num_children = 0

    # for the set of pseudoatoms in this fragment, get the list of pseudoatoms that can react
    allowed_atoms = list()
    for atom in p_atoms:
        atom = int(atom)
        allowed_atoms.extend(brics_mating_rules[atom])
    allowed_atoms = set(allowed_atoms)

    # default settings for max_pseudoatoms assigned here
    # careful changing things, everything is more or less in a delicate balance
    max_pseudoatoms = round(missing_piece_fc * permissivity * user_num_pseudo_atoms) - 1.0

    # get the mate fragments (first pass)
    mate_mols = get_mate_fragments(db,
                                   missing_piece_fc=missing_piece_fc,
                                   missing_piece_len=missing_piece_len,
                                   permissivity=permissivity,
                                   max_pseudoatoms=max_pseudoatoms,
                                   allowed_atoms=allowed_atoms,
                                   max_num_children=max_num_children)

    # get the generator that will produce the children
    children = react(user_frag_mol, mate_mols)

    # catch the end of the generator
    done_mating = False

    # loop through generator until correct number of children generated, or until generator craps out
    while (not done_mating) and (num_children < max_num_children):
        try:
            # get the next child
            try:
                child = next(children)
            except StopIteration:
                done_mating = True
                break
            child_smile = Chem.MolToSmiles(child, isomericSmiles=True)
            child = Chem.MolFromSmiles(child_smile)
            _, _, p_atoms = fragment.get_pseudo_atoms(child_smile)
            num_p_atoms = len(p_atoms)

            # within loop values of parameters, to iteratively change as the loop continues
            # hopefully safe against infinite loops
            max_pseudoatoms_tmp = round(max_pseudoatoms / user_num_pseudo_atoms)
            permissivity_tmp = permissivity
            missing_piece_len_tmp = missing_piece_len

            # this part is where any children that still have pseudoatoms will be "completed"
            # basically, increasingly smaller and less complex pieces will be attached, until no pseudoatoms remain
            retries = 0
            while num_p_atoms > 0:
                # if retries > 5:
                # logger.warning(f"Continuing to mate {user_smile} child # {num_children}, "
                #                f"{child_smile}, filling in child, attempt # {retries}:")
                # decrease overall permissivity
                # for the set of pseudoatoms in this fragment, get the list of pseudoatoms that can react
                allowed_atoms_tmp = list()
                for atom in p_atoms:
                    atom = int(atom)
                    allowed_atoms_tmp.extend(brics_mating_rules[atom])
                allowed_atoms_tmp = set(allowed_atoms_tmp)
                max_pseudoatoms_tmp = max_pseudoatoms_tmp - 1
                permissivity_tmp = round(permissivity_tmp * 0.4, 2)
                mate_mols = get_mate_fragments(db,
                                               missing_piece_fc=missing_piece_fc,
                                               missing_piece_len=missing_piece_len_tmp,
                                               permissivity=permissivity_tmp,
                                               max_pseudoatoms=max_pseudoatoms_tmp,
                                               allowed_atoms=allowed_atoms_tmp,
                                               max_num_children=max_num_children)

                # react the original child with a new mate frag
                complete_children = react(child, mate_mols)

                # get the result as a child, in order to mask the original child
                # this allows convergence to a complete child to occur
                try:
                    child = next(complete_children)
                    child_smile = Chem.MolToSmiles(child, isomericSmiles=True)
                    child = Chem.MolFromSmiles(child_smile)
                except Exception as e:  # pylint: disable=broad-except
                    if retries > 5:
                        # logger.warning(f"Couldn't find any children for {child_smile}.\tTrying again.")
                        permissivity_tmp = round(permissivity_tmp * 2, 2)
                    retries += 1
                    if retries < 10:
                        continue
                    else:
                        logger.error(f"Couldn't find any children for {child_smile}. Aborting.")
                        break

                child_smile = Chem.MolToSmiles(child, isomericSmiles=True)

                # recalculate new  estimated missing piece length
                # basically, how much was added since the original fragment (when missing_piece_len was defined)?
                # remove that much from the estimate of missing_piece_len
                new_diff = len(child_smile) - len(user_smile)
                missing_piece_len_tmp = missing_piece_len - new_diff

                # loop condition, if there are still pseudoatoms, go again
                _, _, p_atoms = fragment.get_pseudo_atoms(child_smile)
                num_p_atoms = len(p_atoms)

            # need to check if the child is complete or not
            child_smile = Chem.MolToSmiles(child, isomericSmiles=True)
            _, _, p_atoms = fragment.get_pseudo_atoms(child_smile)
            num_p_atoms = len(p_atoms)
            if num_p_atoms > 0:
                num_children += 1
                continue

            # fix some chemical details of the generated child and store
            child.UpdatePropertyCache(strict=True)
            Chem.SanitizeMol(child)
            finished_children.append(child)
            num_children += 1
            # logger.debug(f"Finished with child {user_smile} child # {num_children}, {child_smile}.")

        except Exception as e:  # pylint: disable=broad-except
            # generator exhausted, or some other failure
            # (shouldn't matter much, just leave the loop, there will be others)
            logger.error(f"Couldn't mate {user_smile} for some reason.")
            logger.error(f"Error: {e}")
            done_mating = True
            # DO NOT RAISE AN ERROR HERE I SWEAR
            # LOOP HAS TO CONTINUE HERE, ERROR IS KNOWN AND I'M NOT WORRIED ABOUT IT

    logger.info(f"Finished Mating {user_smile}. Produced {len(finished_children)} children.")
    # pre-filter the children of this parent
    pre_filter_values = apply_filter(final_limits,
                                     finished_children,
                                     must_have_patterns=must_have_patterns,
                                     must_not_have_patterns=must_not_have_patterns)

    return finished_children, pre_filter_values


def get_mate_fragments(db,
                       missing_piece_fc,
                       missing_piece_len,
                       permissivity,
                       max_pseudoatoms,
                       allowed_atoms,
                       max_num_children):
    """
    for a given set of criteria, get compatible mate fragments
    db = location of the fragment library database
    missing_piece_fc = the estimated FC of the pieces removed from the user parent, to create this user frag
    missing_piece_len = see above but for length
    permissivity = scaling factor that controls how deviant molecules are allowed to be:
                    1~parity to parent, 0~smallest possible additions, >1~allow adding larger/more complex fragments
    max_num_children = the number of children to return for this fragment
    max_pseudoatoms = not user facing, use caution. The starting value for how many pseudoatoms a mate frag is allowed

    :param db:
    :param missing_piece_fc:
    :param missing_piece_len:
    :param permissivity:
    :param max_pseudoatoms:
    :param allowed_atoms:
    :param max_num_children:
    :return:
    """

    # sanity checks
    # ensure that all the input parameters make sense
    # minimum number of pseudoatoms is 1
    if max_pseudoatoms <= 0.0:
        max_pseudoatoms = 1.0

    # minimum smile length cutoff to use is 15 (chosen apropos of nothing)
    # important to have one of these here, as sometimes the estimated length can be negative for a variety of reasons
    if missing_piece_len < 20.0:
        missing_piece_len = 20.0

    # if the missing piece FC is zero, and we are not yet at the "base case" of missing_piece_len == 0
    # and we are not allowing any more pseudoatoms, then use the missing_piece_len as a rough
    # metric to allow more pseudoatoms, and increase the allowed FC
    # this will ensure that if we start with a very small piece and attach a small piece, then we are more likely
    # to attach multiple pieces in order to try to get back the rough size of the user parent
    if missing_piece_fc == 0.0 and missing_piece_len > 20.0 and max_pseudoatoms == 1.0:
        max_pseudoatoms = round(missing_piece_len / 10)
        missing_piece_fc = (max_pseudoatoms - 1.0)

    # load the fragment library
    if not os.path.exists(db):
        raise Exception("Database file not found: " + db)
    lib = SqliteDatabase(db)
    Fragment, _, _, _ = lib_read(lib)
    with lib:
        # define a length cutoff dynamically, with a minimum
        length_cutoff = ((missing_piece_len * 1.5) / (0.5 * max_pseudoatoms)) * permissivity
        if length_cutoff < 20.0:
            length_cutoff = 20.0

        # this is the part where we avoid the "ORDER BY RANDOM() LIMIT" query by using the FragmentIndex.
        # It saves quite a bit of time, at a moderate cost of memory.
        # returns an iterator rather than a list, so we only fetch as many fragments as we need.
        fx = frag_index(db)
        lim = max_num_children * 100
        fids = fx.mate_query(
            permissivity, max_pseudoatoms,
            missing_piece_fc, missing_piece_len,
            length_cutoff, allowed_atoms,
            limit=lim
        )

        def mate_mols():
            for fid in fids:
                my_mate = Fragment.select().where(Fragment.id == fid)[0]
                mol = Chem.MolFromSmiles(my_mate.smile)
                yield mol

        fast_iterator = mate_mols()
        # Don't close db when we return fast iterator, it will be closed when the iterator closes.
        return fast_iterator



# coding=utf-8
#
# Provide an index into the fragment library that finds fragments to mate with much faster than
# a database query. The problem with the database query is the "ORDER BY RANDOM() LIMIT" clause,
# which steadfastly resists query optimization.
#
# script by Andreas Windemuth

import os
import pickle
from typing import Set
import numpy as np
from peewee import SqliteDatabase
import logging as logger
from .util import lazy
from .fragment_caches import percent, CacheL2
from .fragment_queries import BackupQuery1, BackupQuery2, \
    BackupQuery3, BackupQuery4, MateQuery, FragArray, MateQueryParams
from .lib_read import lib_read

# This defines the backup query cascade. It should be possible to add and remove queries or rearrange them,
#   except after that comparison with the old mate query will no longer be possible after that
query_cascade = [MateQuery(), BackupQuery1(), BackupQuery2(), BackupQuery3(), BackupQuery4()]

# Singletons for fragment index
frag_index_cache = {}


def frag_index(db: str):
    global frag_index_cache  # pylint: disable=global-statement
    name = os.path.splitext(db)[0]
    if name not in frag_index_cache:
        fx = FragmentIndex.by_name(name)
        frag_index_cache[name] = fx
    else:
        fx = frag_index_cache[name]
    return fx


class FragmentIndex(object):
    """
    An index into the fragment base to make mate queries more efficient.
    """

    def __init__(self, frags: FragArray, checking: bool = False):
        self.id_map_value = {}
        self.fragments = frags
        self.id_map_calc = None
        if np.any(frags.id == 0):
            (ix,) = np.where(frags.id == 0)
            raise Exception("Missing fragments: %d " % len(ix) + str(ix))

        # set the cache up
        self.clear_cache()

    @classmethod
    def from_db(cls, db_name: str, limit: int = 0):
        """Read all the fragments from a fragment database"""
        lib = SqliteDatabase(db_name)
        Fragment, _, PseudoAtoms, _ = lib_read(lib)
        lib.connect()
        logger.info("Get count...")
        n_frags = len(Fragment.select())
        if limit > 0:
            n_frags = min(n_frags, limit)
        logger.info("  %d fragments. Loading...." % n_frags)
        fragments = FragArray.zeros(n_frags)
        idmap = {}
        n = 0
        for pa in PseudoAtoms.select().join(Fragment).iterator():
            frag = pa.frag
            ident = int(frag.id)
            k = int(pa.pseudo_atom)
            if ident in idmap:
                i = idmap[ident]
            else:
                i = len(idmap)
                if limit > 0 and i >= limit: break
                idmap[ident] = i
                fragments.id[i] = ident
                fragments.slen[i] = len(frag.smile)
                fragments.coeff[i] = int(frag.frag_coeff)
                fragments.npseudo[i] = int(frag.num_pseudo_atoms)
                if i % 100000 == 0:
                    logger.info(
                        "  %d " % n + percent(i, n_frags) + "  " + "%7d %7d %2d %8d  " % (ident, i, k, int(2 ** k)))
            fragments.pseudo[i] |= int(2 ** k)
            n += 1
        nf = len(idmap)
        flib = FragmentIndex(fragments.truncate(nf))
        flib.id_map_calc = idmap
        logger.info("Read %d/%d fragments, %d pseudo atoms." % (nf, n_frags, n))
        return flib

    @classmethod
    def by_name(cls, name: str, limit: int = 0):
        """Read all the fragments from a binary file, or create on from the fragment database"""
        if limit > 0:
            mod = "-%d" % limit
        else:
            mod = ""
        dbn = name + ".db"
        bfn = name + mod + ".bin"
        if os.path.exists(bfn):
            logger.info("Reading fragments from " + bfn)
            with open(bfn, 'rb') as f:
                fragments = pickle.load(f)
            return FragmentIndex(fragments)
        else:
            logger.info("Loading fragments from " + dbn)
            flib = FragmentIndex.from_db(dbn, limit)
            logger.info("Saving fragments to " + bfn)
            with open(bfn, 'wb') as f:
                pickle.dump(flib.fragments, f, protocol=pickle.HIGHEST_PROTOCOL)
            return flib

    @lazy
    def id_map(self):
        """A mapping to get the index of a fragment from its id"""
        if self.id_map_calc is None:
            self.id_map_value = {}
            for i, ident in enumerate(self.fragments.id):
                self.id_map_calc[ident] = i
        return self.id_map_calc

    def mate_query(self,
                   permissivity: float, max_pseudoatoms: int,
                   missing_piece_fc: int, missing_piece_len,
                   length_cutoff: int, allowed_atoms: Set[int],
                   limit: int
                   ):
        """
        Efficient mate query for deriver.
        :param permissivity:
        :param max_pseudoatoms:
        :param missing_piece_fc:
        :param missing_piece_len:
        :param length_cutoff:
        :param allowed_atoms:
        :param limit:
        :return:  np.array of fragment ids
        """

        # Compute the bit mask for the allowed pseudo atoms
        mask = 0
        for a in allowed_atoms:
            mask |= 2 ** a  # every pseudoatom type corresponds to a position in the bit vector

        params = MateQueryParams(
            permissivity, max_pseudoatoms,
            missing_piece_fc, missing_piece_len,
            length_cutoff, mask
        )

        def cascade(qs) -> np.ndarray:
            ixl: np.ndarray = None
            for k in range(len(qs)):
                ixl = qs[k].select(params)
                if len(ixl) > 0: break
            return ixl

        ix = cascade(self.queries)

        # Find random subset
        np.random.shuffle(ix)
        if limit > 0:
            return self.fragments.id[ix[:limit]]
        else:
            return self.fragments.id[ix]

    def clear_cache(self):
        """clear the cache"""
        self.queries = [CacheL2(self.fragments, q) for q in query_cascade]



"""
Written by Jan H. Jensen 2018, copied 04/2020
"""
from rdkit.Chem import AllChem

import random
import numpy as np
from .jensen_crossover import ring_OK

from rdkit import Chem, rdBase
rdBase.DisableLog('rdApp.error')


def delete_atom():
    choices = ['[*:1]~[D1]>>[*:1]', '[*:1]~[D2]~[*:2]>>[*:1]-[*:2]',
               '[*:1]~[D3](~[*;!H0:2])~[*:3]>>[*:1]-[*:2]-[*:3]',
               '[*:1]~[D4](~[*;!H0:2])(~[*;!H0:3])~[*:4]>>[*:1]-[*:2]-[*:3]-[*:4]',
               '[*:1]~[D4](~[*;!H0;!H1:2])(~[*:3])~[*:4]>>[*:1]-[*:2](-[*:3])-[*:4]']
    p = [0.25, 0.25, 0.25, 0.1875, 0.0625]
  
    return np.random.choice(choices, p=p)


def append_atom():
    choices = [['single',['C','N','O','F','S','Cl','Br'],7*[1.0/7.0]],
               ['double',['C','N','O'],3*[1.0/3.0]],
               ['triple',['C','N'],2*[1.0/2.0]] ]
    p_BO = [0.60, 0.35, 0.05]

    index = np.random.choice(list(range(3)), p=p_BO)

    BO, atom_list, p = choices[index]
    new_atom = np.random.choice(atom_list, p=p)

    if BO == 'single':
      rxn_smarts = '[*;!H0:1]>>[*:1]X'.replace('X','-'+new_atom)
    if BO == 'double':
      rxn_smarts = '[*;!H0;!H1:1]>>[*:1]X'.replace('X','='+new_atom)
    if BO == 'triple':
      rxn_smarts = '[*;H3:1]>>[*:1]X'.replace('X','#'+new_atom)

    return rxn_smarts


def insert_atom():
    choices = [['single',['C','N','O','S'],4*[1.0/4.0]],
               ['double',['C','N'],2*[1.0/2.0]],
               ['triple',['C'],[1.0]] ]
    p_BO = [0.60,0.35,0.05]

    index = np.random.choice(list(range(3)), p=p_BO)

    BO, atom_list, p = choices[index]
    new_atom = np.random.choice(atom_list, p=p)

    if BO == 'single':
      rxn_smarts = '[*:1]~[*:2]>>[*:1]X[*:2]'.replace('X',new_atom)
    if BO == 'double':
      rxn_smarts = '[*;!H0:1]~[*:2]>>[*:1]=X-[*:2]'.replace('X',new_atom)
    if BO == 'triple':
      rxn_smarts = '[*;!R;!H1;!H0:1]~[*:2]>>[*:1]#X-[*:2]'.replace('X',new_atom)

    return rxn_smarts


def change_bond_order():
    choices = ['[*:1]!-[*:2]>>[*:1]-[*:2]','[*;!H0:1]-[*;!H0:2]>>[*:1]=[*:2]',
               '[*:1]#[*:2]>>[*:1]=[*:2]','[*;!R;!H1;!H0:1]~[*:2]>>[*:1]#[*:2]']
    p = [0.45,0.45,0.05,0.05]

    return np.random.choice(choices, p=p)


def delete_cyclic_bond():
    return '[*:1]@[*:2]>>([*:1].[*:2])'


def add_ring():
    choices = ['[*;!r;!H0:1]~[*;!r:2]~[*;!r;!H0:3]>>[*:1]1~[*:2]~[*:3]1',
               '[*;!r;!H0:1]~[*!r:2]~[*!r:3]~[*;!r;!H0:4]>>[*:1]1~[*:2]~[*:3]~[*:4]1',
               '[*;!r;!H0:1]~[*!r:2]~[*:3]~[*:4]~[*;!r;!H0:5]>>[*:1]1~[*:2]~[*:3]~[*:4]~[*:5]1',
               '[*;!r;!H0:1]~[*!r:2]~[*:3]~[*:4]~[*!r:5]~[*;!r;!H0:6]>>[*:1]1~[*:2]~[*:3]~[*:4]~[*:5]~[*:6]1']
    p = [0.05,0.05,0.45,0.45]

    return np.random.choice(choices, p=p)


def change_atom(mol):
    choices = ['#6','#7','#8','#9','#16','#17','#35']
    random.shuffle(choices)
    X = ""
    for choice in choices:
        if choice in Chem.MolToSmarts(mol):
            X = choice
            break
    random.shuffle(choices)
    Y = ""
    for choice in choices:
        if choice != X:
            Y = choice
            break

    if Y == "" or X == "":
        return None
    return '[X:1]>>[Y:1]'.replace('X',X).replace('Y',Y)


def reactor(mol):

    ia = lambda x: insert_atom()
    cbo = lambda x: change_bond_order()
    dcb = lambda x: delete_cyclic_bond()
    ar = lambda x: add_ring()
    da = lambda x: delete_atom()
    ca = lambda x: change_atom(x)
    aa = lambda x: append_atom()

    rxn_func_list = [ia, cbo, dcb, ar, da, ca, aa]
    random.shuffle(rxn_func_list)

    for rxn_func in rxn_func_list:
        rxn_smarts = []
        for _ in range(10):
            rxn_smarts.append(rxn_func(mol))
        rxn_smarts = list(set(rxn_smarts))
        for smarts in rxn_smarts:
            if smarts is None:
                continue
            rxn = AllChem.ReactionFromSmarts(smarts)
            yield rxn


def mutate(mol, mutation_rate):

    if random.random() > mutation_rate:
        return mol

    Chem.Kekulize(mol, clearAromaticFlags=True)
    my_reactor = reactor(mol)
    for rxn in my_reactor:
        new_mol_trial = rxn.RunReactants((mol,))
        for m in new_mol_trial:
            m = m[0]
            if ring_OK(m):
                return m

    return None



"""
Modified from code by Emilie S. Henault and Jan H. Jensen 2019
"""

from rdkit import Chem
import random
from selfies import encoder, decoder

from rdkit import rdBase
rdBase.DisableLog('rdApp.error')


def cut_point(parent):
    m = random.randint(0, len(parent) - 1)
    return m


def mol2string(mol):
    Chem.Kekulize(mol, clearAromaticFlags=True)
    smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
    return encoder(smiles).split('][')


def string2mol(string):
    string = ']['.join(string)
    if not string.endswith("]"):
        string += "]"
    try:
        smiles = decoder(string)
    except:
        return None
    try:
        mol = Chem.MolFromSmiles(smiles)
        return mol
    except:
        return None


def crossover(parent_a_mol, parent_b_mol):
    parent_a, parent_b = mol2string(parent_a_mol), mol2string(parent_b_mol)

    for _ in range(50):
        cut_point_a = cut_point(parent_a)
        cut_point_b = cut_point(parent_b)
        a1 = parent_a[0:cut_point_a]
        b2 = parent_b[cut_point_b:len(parent_b)]
        child_string = a1 + b2
        child_mol = string2mol(child_string)
        return child_mol

    return None



"""
Script to dynamically get the BRICS mating rules, in case they change/update
"""

# this is the reaction definiations
from rdkit.Chem.BRICS import reactionDefs
from collections import defaultdict

# this is the format I want to store them in, a dict of lists
brics_mating_rules = defaultdict(list)

for reaction_set in reactionDefs:
    for reaction in reaction_set:

        # get the two pseudoatom types
        x, y = reaction[0], reaction[1]

        """
        deal with (read: ignore) their special definition of pseudoatom 7
        this causes some problems later, but ultimately is still simpler to deal with
        we need the value to be an integer for the database
        """
        if x == "7a" or x == "7b":
            x = 7
        if y == "7a" or y == "7b":
            y = 7

        # I really must insist that these be integers
        x = int(x)
        y = int(y)

        # store symmetrically
        brics_mating_rules[x].append(y)
        brics_mating_rules[y].append(x)

# remove redundant entries in each list
for key in brics_mating_rules:
    tmp_set = set(brics_mating_rules[key])
    brics_mating_rules[key] = list(tmp_set)



# coding=utf-8
#
# Alternate implementations of the cascading mate queries in mate.py, to go with the FragmentIndex..
#
# code by Andreas Windemuth

from typing import NamedTuple
import numpy as np
from .util import Vec


class MateQueryParams(NamedTuple):
    """Collection of all the parameters a query may depend on"""
    permissivity: float
    max_pseudoatoms: int
    missing_piece_fc: int
    missing_piece_len: int
    length_cutoff: int
    mask: int  # bit encoded version of allowed_atoms


class FragArray(NamedTuple):
    """An array of all the information about fragments needed to support mate queries.
       Note that this does not include SMILES, which need to still be read from the database.
       We could/should add smiles, but need to watch memory consumption.
    """
    id: Vec[int]  # fragment id
    slen: Vec[int]  # SMILES length
    coeff: Vec[int]  # frag_coeff (whatever it is...)
    npseudo: Vec[int]  # number of pseudo atoms
    pseudo: Vec[int]  # bitset of pseudo atoms

    @classmethod
    def zeros(cls, n):
        """Provide an empty array of a certain length"""

        def idim(): return np.zeros(n, dtype=int)

        return FragArray(id=idim(), slen=idim(), coeff=idim(), npseudo=idim(), pseudo=idim())

    def truncate(self, n):
        """Apply an index or index range.
           Can't call it __getitem__ as this conflicts with the tuple API and leads to infinite recursion
        """
        return FragArray(
            id=self.id[:n], slen=self.slen[:n], coeff=self.coeff[:n], npseudo=self.npseudo[:n], pseudo=self.pseudo[:n]
        )


# These are the cascading queries, in order of greatest stringency (except one seems superfluous?)

class MateQuery(object):
    def key(self, params: MateQueryParams) -> str:
        return "%d %d %d %d %d %d" % (
            int(params.max_pseudoatoms), int(params.max_pseudoatoms / 3.0),
            int(params.permissivity * params.missing_piece_fc), int(params.max_pseudoatoms / 3.0),
            params.length_cutoff, int(params.length_cutoff / 3.0)
        )

    def clause(self, fragments: FragArray, params: MateQueryParams):
        # logger.info("    npa: %d - %d" % (int(params.max_pseudoatoms / 3.0), int(params.max_pseudoatoms)))
        # logger.info("  coeff: %d - %d" % (int((params.permissivity / 3.0) * params.missing_piece_fc), int(params.permissivity * params.missing_piece_fc)))
        # logger.info("  smile: %d - %d" % (int(params.length_cutoff / 3.0), int(params.length_cutoff)))
        return (
                (fragments.npseudo <= int(params.max_pseudoatoms)) &
                (fragments.npseudo >= int(params.max_pseudoatoms / 3.0)) &
                (fragments.coeff <= int(params.permissivity * params.missing_piece_fc)) &
                (fragments.coeff >= int((params.permissivity / 3.0) * params.missing_piece_fc)) &
                (fragments.slen <= params.length_cutoff) &
                (fragments.slen >= int(params.length_cutoff / 3.0))
        )


class BackupQuery1(object):
    def key(self, params: MateQueryParams) -> str:
        return "%d %d %d" % (
            int(params.max_pseudoatoms),
            int(params.permissivity * params.missing_piece_fc),
            params.length_cutoff
        )

    def clause(self, fragments: FragArray, params: MateQueryParams):
        return (
                (fragments.npseudo <= int(params.max_pseudoatoms)) &
                (fragments.coeff <= int(params.permissivity * params.missing_piece_fc)) &
                (fragments.slen <= params.length_cutoff)
        )


def alt_length_cutoff(params: MateQueryParams):
    return max(20.0, (params.missing_piece_len * 1.5) / params.max_pseudoatoms)


class BackupQuery2(object):
    def key(self, params: MateQueryParams) -> str:
        return "%d %d %d" % (
            int(params.max_pseudoatoms),
            int(params.permissivity * params.missing_piece_fc),
            alt_length_cutoff(params)
        )

    def clause(self, fragments: FragArray, params: MateQueryParams):
        return (
                (fragments.npseudo <= int(params.max_pseudoatoms)) &
                (fragments.coeff <= int(params.permissivity * params.missing_piece_fc)) &
                (fragments.slen <= alt_length_cutoff(params))
        )


# This query appears to be more strict than the one above, and should therefore never be used ???

class BackupQuery3(object):
    def key(self, params: MateQueryParams) -> str:
        return "%d %d" % (int(params.max_pseudoatoms), alt_length_cutoff(params))

    def clause(self, fragments: FragArray, params: MateQueryParams):
        return (
                (fragments.npseudo <= int(params.max_pseudoatoms)) &
                (fragments.coeff == 0) &
                (fragments.slen <= alt_length_cutoff(params))
        )


class BackupQuery4(object):
    def key(self, params: MateQueryParams) -> str:
        return "single"  # clause does not depend on params

    def clause(self, fragments: FragArray, params: MateQueryParams):
        return (
                (fragments.npseudo == 1) & (fragments.coeff == 0)
        )



from argparse import ArgumentParser
from playhouse.sqlite_ext import SqliteExtDatabase
import logging as logger
from .lib_read import lib_read


def clean(db_loc):
    """
    Some fragments are duplicates, or are not included in the db, but there are still atom and pseudoatom records
    that refer to them. This is not ideal, so this script deletes these orphan records.
    :param db_loc:
    :return:
    """
    # create the database for the output
    db = SqliteExtDatabase(db_loc, pragmas={
        'cache_size': -1024 * 64,  # 64MB page-cache.
        'journal_mode': 'wal',  # Use WAL-mode (you should always use this!).
        'foreign_keys': 0,
        'wal_autocheckpoint': 10,
    })
    # get the models
    Fragment, Heritage, PseudoAtoms, Atoms = lib_read(db)

    Fragment.create_table(safe=True)
    Heritage.create_table(safe=True)
    PseudoAtoms.create_table(safe=True)
    Atoms.create_table(safe=True)

    logger.debug("Trying to clean up the database:")
    logger.debug("Deleting missing ATOM records")
    with db.atomic():
        bad_atoms = Atoms.delete().where(
            (Atoms.frag.not_in(Fragment.select()))
        )
        bad_atoms.execute()

    logger.debug("Deleting missing PSEUDO_ATOM records")
    with db.atomic():
        bad_patoms = PseudoAtoms.delete().where(
            (PseudoAtoms.frag.not_in(Fragment.select()))
        )
        bad_patoms.execute()

    logger.debug("Deleting missing HERITAGE records")
    with db.atomic():
        bad_heritage = Heritage.delete().where(
            (Heritage.frag.not_in(Fragment.select()))
        )
        bad_heritage.execute()

    logger.info("Done.")


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument('-i',
                        dest="input",
                        type=str,
                        help='.db file of molecules to be cleaned.',
                        required=True)

    options = parser.parse_args()
    clean(options.input)



from rdkit import Chem
from selfies import encoder, decoder, get_semantic_robust_alphabet
import re
import random
import logging as logger


ALPHABET = [symbol.strip("[]") for symbol in get_semantic_robust_alphabet()]

# follow these rules for which types of substitution should be allowed
ALLOWED_SUBS = {
    "C": ["N", "O", "H"],
    "=C": ["=N", "N", "O", "=O", "S"],
    "F": ["Cl", "Br", "I", "O", "N", "C", "H"],
    "Cl": ["F", "Br", "I", "O", "N", "C", "H"],
    "Br": ["Cl", "F", "I", "O", "N", "C", "H"],
    "I": ["Cl", "Br", "F", "O", "N", "C", "H"],
    "H": ["C", "O", "N", "S", "=C", "=O", "=S"],
    "O": ["S", "N", "=O", "=N", "C", "=C", "Cl", "F", "Br", "I", "H"],
    "=O": ["=S", "=N", "=C", "O"],
    "N": ["O", "C", "H"],
    "=N": ["=O", "O", "S", "=C", "C"],
    "#N": ["#C"],
    "S": ["O", "N", "C", "=O", "=N", "H"],
    "=S": ["=O", "=N", "=C", "O"],
    "#C": ["#N"]
}


def selfies_substitution(*,
                         parent_smiles: str,
                         n_children: int = 100,
                         mut_rate: float = 0.03,
                         mut_min: int = 1,
                         mut_max: int = 2,
                         max_trials: int = 100
                         ):

    """
    This function takes a parent molecule, and generates a number of children derived from it, via converting
    the parent into SELFIES format and substituting symbols for other symbols, based on primitive chemical rules.
    :param parent_smiles: The smiles of the parent molecule.
    :param n_children: How many children to produce.
    :param mut_rate: How frequent should mutations be? 0.0 to 1.0
    :param mut_min: What is the min number of mutations to allow in a given child, relative to the parent?
    :param mut_max: Same as above but the max.
    :param max_trials: number of attempts to create valid mutations before moving on, useful for pathological selfies
    :return:
    """

    # get the parent mol set up properly with defined aromaticity
    parent_mol = Chem.MolFromSmiles(parent_smiles, sanitize=True)
    Chem.rdmolops.Kekulize(parent_mol)
    parent_smiles = Chem.MolToSmiles(parent_mol, isomericSmiles=True, kekuleSmiles=True)
    logger.info(f"Generating children from: {parent_smiles}")

    children = []  # finished children
    spawns = 0  # counter for children produced
    parent_selfies = encoder(parent_smiles)
    symbols = re.findall(r"[^[]*\[([^]]*)\]", parent_selfies)   # get the SELFIES symbols into a list

    while spawns < n_children:  # try to produce the correct number of children
        muts = 0
        mutations = []  # which parts of the SELFIES to remove
        mut_symbols = symbols.copy()  # don't manipulate the original
        mut_positions = list(range(len(symbols)))  # need the index

        t = 0
        while (muts < mut_min) and (t <= max_trials):
            random.shuffle(mut_positions)  # shuffle the order so that mutations will be random
            for pos in mut_positions:  # try to mutate
                if pos not in mutations:
                    if random.random() <= mut_rate:
                        # ignore special symbols, leave them alone
                        if not (symbols[pos] == "epsilon" or "Branch" in symbols[pos] or "Ring" in symbols[pos]):
                            if symbols[pos] in ALLOWED_SUBS:  # if we have rules for how to handle this symbol
                                mutations.append(pos)  # record which symbol this is
                                muts += 1  # record the intention to mutate
                                if muts == mut_max:
                                    # when we're done, stop looking
                                    break
            t += 1
        if t > max_trials:
            logger.warning(f"Failed to produce any selfies after {max_trials} trials. Returning empty list.")
            return list()

        # for each planned mutation, actually execute it, on the non-shuffled original SELFIES list
        for index in sorted(mutations, reverse=True):
            mut_symbols[index] = random.choice(ALLOWED_SUBS[mut_symbols[index]])

        # convert the new child into a SELFIES (rather than a list)
        child_symbols = [f"[{symb}]" for symb in mut_symbols]
        child = "".join(child_symbols)
        child_smiles = decoder(child)  # get the smiles

        # test that smiles is valid
        try:
            # same as parent, have to have explicit aromaticity
            child_mol = Chem.MolFromSmiles(child_smiles, sanitize=True)
            Chem.rdmolops.Kekulize(child_mol)
            child_smiles = Chem.MolToSmiles(child_mol, isomericSmiles=True, kekuleSmiles=True)
            assert child_mol  # if MolToSmiles fails, it will be a None
            if child_smiles == parent_smiles:  # ignore this child if it's the same as the parent
                continue
        except Exception:  # pylint: disable=broad-except
            logger.warning(f"Produced improper SELFIES. Ignoring and trying again. Details below:")
            logger.warning(f"Child SELFIES: {child}")
            logger.warning(f"Parent SELFIES:{parent_selfies}")
            logger.warning(f"Child SMILES: {child_smiles}")
            logger.warning(f"Parent SMILES: {parent_smiles}")
            continue
        # Every good child deserves fudge
        children.append(child_smiles)
        spawns += 1  # update our counter

    return children


def selfies_insertion(*,
                         parent_smiles: str,
                         n_children: int = 100,
                         mut_rate: float = 0.03,
                         mut_min: int = 1,
                         mut_max: int = 2,
                         max_trials: int = 100
                         ):

    """
    This function takes a parent molecule, and generates a number of children derived from it, via converting
    the parent into SELFIES format and inserting other symbols.
    :param parent_smiles: The smiles of the parent molecule.
    :param n_children: How many children to produce.
    :param mut_rate: How frequent should mutations be? 0.0 to 1.0
    :param mut_min: What is the min number of mutations to allow in a given child, relative to the parent?
    :param mut_max: Same as above but the max.
    :param max_trials: number of attempts to create valid mutations before moving on, useful for pathological selfies
    :return:
    """

    # get the parent mol set up properly with defined aromaticity
    parent_mol = Chem.MolFromSmiles(parent_smiles, sanitize=True)
    Chem.rdmolops.Kekulize(parent_mol)
    parent_smiles = Chem.MolToSmiles(parent_mol, isomericSmiles=True, kekuleSmiles=True)
    logger.info(f"Generating children from: {parent_smiles}")

    children = []  # finished children
    spawns = 0  # counter for children produced
    parent_selfies = encoder(parent_smiles)
    symbols = re.findall(r"[^[]*\[([^]]*)\]", parent_selfies)   # get the SELFIES symbols into a list

    while spawns < n_children:  # try to produce the correct number of children
        muts = 0
        mutations = []  # which parts of the SELFIES to remove
        mut_symbols = symbols.copy()  # don't manipulate the original
        mut_positions = list(range(len(symbols)))  # need the index

        t = 0
        while (muts < mut_min) and (t <= max_trials):
            random.shuffle(mut_positions)  # shuffle the order so that mutations will be random
            for pos in mut_positions:  # try to mutate
                if pos not in mutations:
                    if random.random() <= mut_rate:
                        mutations.append(pos)  # record which symbol this is
                        muts += 1  # record the intention to mutate
                        if muts == mut_max:
                            # when we're done, stop looking
                            break

            t += 1
        if t > max_trials:
            logger.warning(f"Failed to produce any selfies after {max_trials} trials. Returning empty list.")
            return list()

        # for each planned mutation, actually execute it, on the non-shuffled original SELFIES list
        for index in sorted(mutations, reverse=True):
            mut_symbols.insert(index, random.choice(ALPHABET))

        # convert the new child into a SELFIES (rather than a list)
        child_symbols = [f"[{symb}]" for symb in mut_symbols]
        child = "".join(child_symbols)
        child_smiles = decoder(child)  # get the smiles

        # test that smiles is valid
        try:
            # same as parent, have to have explicit aromaticity
            child_mol = Chem.MolFromSmiles(child_smiles, sanitize=True)
            Chem.rdmolops.Kekulize(child_mol)
            child_smiles = Chem.MolToSmiles(child_mol, isomericSmiles=True, kekuleSmiles=True)
            assert child_mol  # if MolToSmiles fails, it will be a None
            if child_smiles == parent_smiles:  # ignore this child if it's the same as the parent
                continue
        except Exception:  # pylint: disable=broad-except
            logger.warning(f"Produced improper SELFIES. Ignoring and trying again. Details below:")
            logger.warning(f"Child SELFIES: {child}")
            logger.warning(f"Parent SELFIES:{parent_selfies}")
            logger.warning(f"Child SMILES: {child_smiles}")
            logger.warning(f"Parent SMILES: {parent_smiles}")
            continue
        # Every good child deserves fudge
        children.append(child_smiles)
        spawns += 1  # update our counter

    return children


def selfies_deletion(*,
                         parent_smiles: str,
                         n_children: int = 100,
                         mut_rate: float = 0.03,
                         mut_min: int = 1,
                         mut_max: int = 2,
                         max_trials: int = 100
                         ):

    """
    This function takes a parent molecule, and generates a number of children derived from it, via converting
    the parent into SELFIES format and deleting symbols.
    :param parent_smiles: The smiles of the parent molecule.
    :param n_children: How many children to produce.
    :param mut_rate: How frequent should mutations be? 0.0 to 1.0
    :param mut_min: What is the min number of mutations to allow in a given child, relative to the parent?
    :param mut_max: Same as above but the max.
    :param max_trials: number of attempts to create valid mutations before moving on, useful for pathological selfies
    :return:
    """

    # get the parent mol set up properly with defined aromaticity
    parent_mol = Chem.MolFromSmiles(parent_smiles, sanitize=True)
    Chem.rdmolops.Kekulize(parent_mol)
    parent_smiles = Chem.MolToSmiles(parent_mol, isomericSmiles=True, kekuleSmiles=True)
    logger.info(f"Generating children from: {parent_smiles}")

    children = []  # finished children
    spawns = 0  # counter for children produced
    parent_selfies = encoder(parent_smiles)
    symbols = re.findall(r"[^[]*\[([^]]*)\]", parent_selfies)   # get the SELFIES symbols into a list

    while spawns < n_children:  # try to produce the correct number of children
        muts = 0
        mutations = []  # which parts of the SELFIES to remove
        mut_symbols = symbols.copy()  # don't manipulate the original
        mut_positions = list(range(len(symbols)))  # need the index

        t = 0
        while (muts < mut_min) and (t <= max_trials):
            random.shuffle(mut_positions)  # shuffle the order so that mutations will be random
            for pos in mut_positions:  # try to mutate
                if pos not in mutations:
                    if random.random() <= mut_rate:
                        mutations.append(pos)  # record which symbol this is
                        muts += 1  # record the intention to mutate
                        if muts == mut_max:
                            # when we're done, stop looking
                            break

            t += 1
        if t > max_trials:
            logger.warning(f"Failed to produce any selfies after {max_trials} trials. Returning empty list.")
            return list()

        # for each planned mutation, actually execute it, on the non-shuffled original SELFIES list
        for index in sorted(mutations, reverse=True):
            del mut_symbols[index]

        # convert the new child into a SELFIES (rather than a list)
        child_symbols = [f"[{symb}]" for symb in mut_symbols]
        child = "".join(child_symbols)
        child_smiles = decoder(child)  # get the smiles

        # test that smiles is valid
        try:
            # same as parent, have to have explicit aromaticity
            child_mol = Chem.MolFromSmiles(child_smiles, sanitize=True)
            Chem.rdmolops.Kekulize(child_mol)
            child_smiles = Chem.MolToSmiles(child_mol, isomericSmiles=True, kekuleSmiles=True)
            assert child_mol  # if MolToSmiles fails, it will be a None
            if child_smiles == parent_smiles:  # ignore this child if it's the same as the parent
                continue
        except Exception:  # pylint: disable=broad-except
            logger.warning(f"Produced improper SELFIES. Ignoring and trying again. Details below:")
            logger.warning(f"Child SELFIES: {child}")
            logger.warning(f"Parent SELFIES:{parent_selfies}")
            logger.warning(f"Child SMILES: {child_smiles}")
            logger.warning(f"Parent SMILES: {parent_smiles}")
            continue
        # Every good child deserves fudge
        children.append(child_smiles)
        spawns += 1  # update our counter

    return children


def random_selfies_generator(*, n_symbols: int = 100):

    rand_gen_alphabet = (list(ALLOWED_SUBS.keys()) * 2) + (["C"] * 50) + ALPHABET

    while True:
        symbols = [random.choice(rand_gen_alphabet) for i in range(n_symbols)]
        # print(symbols)
        # convert the new child into a SELFIES (rather than a list)
        child_symbols = [f"[{symb}]" for symb in symbols]
        child = "".join(child_symbols)
        child_smiles = decoder(child)  # get the smiles
        if len(child_smiles) < 10:
            continue
        elif Chem.MolFromSmiles(child_smiles) is None:
            continue
        else:
            yield child_smiles


def selfies_scanner(*, parent_smiles: str):
    # get the parent mol set up properly with defined aromaticity
    parent_mol = Chem.MolFromSmiles(parent_smiles, sanitize=True)
    Chem.rdmolops.Kekulize(parent_mol)
    parent_smiles = Chem.MolToSmiles(parent_mol, isomericSmiles=True, kekuleSmiles=True)
    logger.info(f"Generating children from: {parent_smiles}")

    children = []  # finished children
    spawns = 0  # counter for children produced
    parent_selfies = encoder(parent_smiles)
    symbols = re.findall(r"[^[]*\[([^]]*)\]", parent_selfies)  # get the SELFIES symbols into a list

    for i, symb in enumerate(symbols):
        if not (symb == "epsilon" or "Branch" in symb or "Ring" in symb):
            if symb in ALLOWED_SUBS:  # if we have rules for how to handle this symbol
                for replacement in ALLOWED_SUBS[symb]:
                    mut_symbols = symbols.copy()  # don't manipulate the original
                    mut_symbols[i] = replacement
                    child_symbols = [f"[{symb}]" for symb in mut_symbols]
                    child = "".join(child_symbols)
                    child_smiles = decoder(child)  # get the smiles
                    # test that smiles is valid
                    try:
                        # same as parent, have to have explicit aromaticity
                        child_mol = Chem.MolFromSmiles(child_smiles, sanitize=True)
                        Chem.rdmolops.Kekulize(child_mol)
                        child_smiles = Chem.MolToSmiles(child_mol, isomericSmiles=True, kekuleSmiles=True)
                        assert child_mol  # if MolToSmiles fails, it will be a None
                        if child_smiles == parent_smiles:  # ignore this child if it's the same as the parent
                            continue
                    except Exception:  # pylint: disable=broad-except
                        logger.warning(f"Produced improper SELFIES. Ignoring and trying again. Details below:")
                        logger.warning(f"Child SELFIES: {child}")
                        logger.warning(f"Parent SELFIES:{parent_selfies}")
                        logger.warning(f"Child SMILES: {child_smiles}")
                        logger.warning(f"Parent SMILES: {parent_smiles}")
                        continue
                    # Every good child deserves fudge
                    children.append(child_smiles)
                    spawns += 1  # update our counter

    return children



from .api import Deriver


def test_deriver():

    deriver = Deriver()
    deriver.set_seeds(['CN(CC(O)COc1nccc(-c2ccc3c(c2)SCC(=O)N3)n1)CC(O)Cn1c(=O)c2ccccc2n(C)c1=O',
                       'CN(CC(=O)Nc1onc2ccc(-c3ccnc(N)n3)cc12)CC(O)Cn1c(=O)c2ccccc2n(C)c1=O',
                       'COc1ccc2c(c1)c(NC(=O)CN(C)CCNc1nccc(-c3c[nH][nH]c3=O)n1)cc(=O)n2C',
                       'CN(CC(=O)Nc1onc2ccc(-c3ccnc(N)n3)cc12)CC(O)CC1(O)COc2cc(O)cc(O)c2C1=O',
                       'COc1cccc(NC(=O)CN(C)CC(=O)c2c(N)n(-c3nc(N4CCN(C)CC4)nc4ccccc34)c(=O)n(C)c2=O)c1',
                       'COc1cc(N2CCN(C)CC2)ccc1Nc1nc(N)nn1C(=O)NCCC(N)C(=O)NNC(=O)c1cc2ccc(-c3ccnc(N)n3)cc2[nH]1',
                       'COc1cc(N2CCN(C)CC2)ccc1Nc1nc(N)nn1C(=O)Nc1ccncc1-c1nnc(C2(O)CCC(=O)N2)o1',
                       'Nc1nccc(-c2ccc3noc(NC(=O)COc4ccc(O)c(S(=O)(=O)N5C=CC(=O)C(O)C5)c4)c3c2)n1',
    ])
    deriver.enable_and_expand_filter()
    selfies_gb, _ = deriver.derive_gb(100, kind='selfies')
    smiles_gb, _ = deriver.derive_gb(100, kind='smiles')
    selfies, _ = deriver.derive_selfies(100)
    assert len(selfies_gb) > 2
    assert len(smiles_gb) > 2
    assert len(selfies) > 2
    assert deriver.data.all_good_selfies_gb_children
    assert deriver.data.all_good_smiles_gb_children



""""Fragment submodule, for fragmenting and calculating fragmentation coefficients."""

import time
from argparse import ArgumentParser
from collections import deque
from playhouse.sqlite_ext import SqliteExtDatabase
from .lib_read import lib_read
from peewee import chunked
from rdkit import RDLogger
from rdkit import Chem
from rdkit.Chem import BRICS
import logging as logger
from copy import deepcopy
import re
from .clean_frag_db import clean

# set up rdkit logger to ignore WARNINGS (it was so irritating (and possibly responsible for a slight slowdown))
lg = RDLogger.logger()
lg.setLevel(RDLogger.ERROR)

# these atoms are allowed in the fragments, other atoms are not drug-like enough for our purposes
# TODO: could be made into an argument/parameter to be user-provided
allowed_atoms = [
    "C", "N", "S", "O", "P",
    "F", "Br", "Cl", "Na", "K",
    "Mg", "Ca", "B", "Si", "I",
    "Fe", "H", "*"
]


# count atoms (a thread-safe counter that can provide unique IDs) (was more important when multithreaded)
class Counter(object):
    def __init__(self, counter=0):
        self.val = counter

    def increment(self, n=1):
        val = deepcopy(self.val)
        self.val = val + n
        return val


ID_COUNTER = Counter()


class RecomposerMol(object):
    """
    This is a graph-based description of a molecule, which allows for better handling of stereochemistry when
    breaking and reattaching bonds. Otherwise, stereochemistry gets pretty junked up. This also permits fragments
    to be built 'up' from the smallest pieces, producing more fragments, faster, than using the builtin rdkit
    methods. This function based off draft code developed by Harsh Patel @harsh6646
    """
    atomCounter = None
    pseudoIndex = {}  # where are the pseudoatoms if any
    atoms = {}  # same but for atoms
    fragments = {}  # here are all the fragments, keys are FC values
    init_mol = None  # rdkit mol object of this molecule
    edges = []  # basically where are all the bonds
    fc0_index = {}  # given a specific pseudoatom, which fc0 fragment has it
    frag_cache = []  # previously seen fragments (no need to build up from them twice) (can't be a set, dicts inside)
    brics_legend = {}  # which BRICS pseudoatom type is each pseudoatom, needed to rebuild at the end

    def __eq__(self, item):
        if isinstance(item, RecomposerMol):
            return self.pseudoIndex == item.pseudoIndex
        else:
            return self.pseudoIndex == item

    def __hash__(self):
        return self.pseudoIndex

    # track atoms
    class Atom():
        index = None
        atom_num = None
        isotope = None
        chiralTag = None
        aromaticTag = None
        bonds = None
        explicitHs = None

        def __init__(self, *, index:int, atom_num:int, isotope:int,
                     chiralTag:Chem.rdchem.ChiralType, explicitHs:int, bonds:[],
                     aromaticTag: bool):
            self.index = index
            self.atom_num = atom_num
            self.isotope = isotope
            self.chiralTag = chiralTag
            self.explicitHs = explicitHs
            self.aromaticTag = aromaticTag
            self.bonds = bonds

    # track bonds
    class Bond():
        bondType = None
        beginAtomIdx = None
        endAtomIdx = None
        # may want to add bond dir and stereo if needed

        def __init__(self, bondType, beginAtomIdx, endAtomIdx):
            self.bondType = bondType
            self.beginAtomIdx = beginAtomIdx
            self.endAtomIdx = endAtomIdx

    @staticmethod
    def fromMol(mol):
        """
        Build a RecomposerMol object from an rdkit Mol object
        :param mol: rdkit Mol object
        :return: a RecomposerMol object
        """
        newMol = RecomposerMol()
        atoms = {}
        pseudoIndex = {}
        # add info for each atom in the mol into the atom dict
        for atom in mol.GetAtoms():
            idx = atom.GetIdx()
            bonds = atom.GetBonds()
            aromaticTag = atom.GetIsAromatic()
            newBonds = [newMol.Bond(bond.GetBondType(), bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()) for bond in bonds]
            newAtom = newMol.Atom(index=idx, atom_num=atom.GetAtomicNum(),
                                         isotope=atom.GetIsotope(), chiralTag=atom.GetChiralTag(),
                                         explicitHs=atom.GetNumExplicitHs(), bonds=newBonds,
                                         aromaticTag=aromaticTag)
            atoms[idx] = newAtom
            if atom.GetAtomicNum() == 0:  # if it's a pseudoatom
                pseudoIndex[atom.GetIsotope()] = idx
        newMol.init_mol = mol
        newMol.atoms = atoms  # setting these like this prevents reference errors
        newMol.atomCounter = Counter(mol.GetNumAtoms())
        newMol.pseudoIndex = pseudoIndex  # likewise here
        newMol.frag_cache = [{}]  # the parent mol always has an empty pseudoindex
        return newMol

    def addPseudoAtom(self, isotope):
        new_bonds = []
        # get a counter to use as an id
        idx = self.atomCounter.increment()
        new_atom = self.Atom(index=idx, atom_num=0, isotope=isotope,
                            chiralTag=Chem.rdchem.ChiralType.CHI_UNSPECIFIED, explicitHs=0,
                            bonds=new_bonds, aromaticTag=False)
        self.atoms[idx] = new_atom
        self.pseudoIndex[isotope] = idx  # track in order to replace at the end
        return idx

    def breakAndReplace(self, beginIdx, endIdx, beginValue, endValue):
        """
        Replace a bond with two bonds to pseudoatoms. Edit in place.
        :param beginIdx: start atom
        :param endIdx:  end atom
        :param beginValue:  pseudoatom type for start atom
        :param endValue: pseudoatom type for end atom
        :return:
        """
        # just edit in place
        begin_prime = self.addPseudoAtom(beginValue)
        end_prime = self.addPseudoAtom(endValue)

        # replace first bond to bond with pseudoatom
        for i, bond in enumerate(self.atoms[beginIdx].bonds):
            # what follows is a bunch of checks to detect how the rdkit object bond and atom order matches the desired
            if (bond.beginAtomIdx, bond.endAtomIdx) == (beginIdx, endIdx):
                self.atoms[beginIdx].bonds[i].endAtomIdx = begin_prime
                new_bonds = [self.Bond(bondType=bond.bondType, beginAtomIdx=beginIdx, endAtomIdx=begin_prime)]
                self.atoms[begin_prime].bonds = new_bonds
                break
            elif (bond.beginAtomIdx, bond.endAtomIdx) == (endIdx, beginIdx):
                self.atoms[beginIdx].bonds[i].beginAtomIdx = begin_prime
                new_bonds = [self.Bond(bondType=bond.bondType, beginAtomIdx=begin_prime, endAtomIdx=beginIdx)]
                self.atoms[begin_prime].bonds = new_bonds
                break
        # replace second bond with bond with pseudoatom
        for i, bond in enumerate(self.atoms[endIdx].bonds):
            if (bond.beginAtomIdx, bond.endAtomIdx) == (beginIdx, endIdx):
                self.atoms[endIdx].bonds[i].beginAtomIdx = end_prime
                new_bonds = [self.Bond(bondType=bond.bondType, beginAtomIdx=end_prime, endAtomIdx=endIdx)]
                self.atoms[end_prime].bonds = new_bonds
                break
            elif (bond.beginAtomIdx, bond.endAtomIdx) == (endIdx, beginIdx):
                self.atoms[endIdx].bonds[i].endAtomIdx = end_prime
                new_bonds = [self.Bond(bondType=bond.bondType, beginAtomIdx=endIdx, endAtomIdx=end_prime)]
                self.atoms[end_prime].bonds = new_bonds
                break

    def get_fc0_frags(self):
        """
        This function breaks every bond and enumerates all the FC0 fragments. This is a needed first step, since these
        are used to build up all the more complex fragments.
        :return:
        """
        fc0_frags = []
        brics = BRICS.FindBRICSBonds(self.init_mol)
        brics_counter = Counter(counter=100)  # counter starts at 100 since it's not going to be in use as a p.atom
        edges = []
        for bond, brics_type in brics:
            i, j = bond
            x, y = brics_type
            pseudo_i = brics_counter.increment()
            pseudo_j = brics_counter.increment()  # this is always done in pairs so that eg. 100 and 101 go together
            self.brics_legend[pseudo_i] = int(x)  # remember to store the BRICS typing, we need it later
            self.brics_legend[pseudo_j] = int(y)
            edges.append((pseudo_i, pseudo_j))  # keep a record of this bond
            self.breakAndReplace(i, j, pseudo_i, pseudo_j)
        self.edges = edges

        # this while loop finds parts of a graph that are no longer connected, meaning FC0 fragments in this case
        untouched = set(self.atoms.keys())  # keep a list of atoms we haven't investigated yet
        while untouched:  # if there are still atoms we haven't processed
            currSet = set()
            currSet.add(untouched.pop())
            newMol = RecomposerMol()
            newAtoms = {}
            newPseudoAtoms = {}
            new_brics_legend = {}
            new_brics_legend.update(self.brics_legend)
            while currSet:
                # perform graph search
                currAtom = self.atoms[currSet.pop()]
                for bond in currAtom.bonds:
                    idx = None
                    if currAtom.index == bond.endAtomIdx:
                        idx = bond.beginAtomIdx
                    elif currAtom.index == bond.beginAtomIdx:
                        idx = bond.endAtomIdx
                    # add connected atoms to curr mol atom set
                    if idx in untouched:
                        currSet.add(idx)
                        untouched.remove(idx)
                    else:
                        continue
                newAtoms[currAtom.index] = self._copyAtom(currAtom)  # watch out for reference errors!
                if currAtom.atom_num == 0:
                    newPseudoAtoms[currAtom.isotope] = currAtom.index
            newMol.atoms = newAtoms
            newMol.pseudoIndex = newPseudoAtoms
            newMol.brics_legend = new_brics_legend
            fc0_frags.append(newMol)
            self.frag_cache.append(newMol)
        # now need to index the fragments to make it easier to bond them
        fc0_index = {}
        for i, frag in enumerate(fc0_frags):
            for pseudo_atom in frag.pseudoIndex.keys():
                fc0_index[pseudo_atom] = i  # each FC0 fragment should be accessible by pseudoatom
        self.fc0_index = fc0_index
        self.fragments[0] = fc0_frags

    def react(self, *, frag1, frag2, p_atom1, p_atom2):
        """
        basically the opposite of breaking the bonds. Remove the pseudoatoms and put a bond there.
        :param frag1: RecomposerMol
        :param frag2: RecomposerMol
        :param p_atom1: the pseudoatom on fragment 1
        :param p_atom2: the pseudoatom on fragment 2
        :return: a new RecomposerMol
        """

        # merge the two mols
        new_mol = RecomposerMol()
        newAtoms = {}
        new_pi = {}
        new_bl = {}
        new_bl.update(frag1.brics_legend)  # the brics legend is so useful, keep it
        for atom_idx in frag1.atoms:
            newAtoms[atom_idx] = frag1._copyAtom(frag1.atoms[atom_idx])
        for atom_idx in frag2.atoms:
            newAtoms[atom_idx] = frag2._copyAtom(frag2.atoms[atom_idx])

        new_mol.atoms = newAtoms
        pi1 = deepcopy(frag1.pseudoIndex)
        new_pi.update(pi1)
        pi2 = deepcopy(frag2.pseudoIndex)
        new_pi.update(pi2)

        # new mol is ready, now bond it and remove the pseudoatoms
        # get the real atoms that are to be attached
        p_a_idx = new_pi[p_atom1]
        p_b_idx = new_pi[p_atom2]
        a_idx = None
        b_idx = None

        # need to check each end of both bonds to figure out which is the right atom to replace
        for bond in new_mol.atoms[p_a_idx].bonds:  # should only ever be one bond for any pseudoatom
            if bond.beginAtomIdx == p_a_idx:
                a_idx = bond.endAtomIdx
            else:
                a_idx = bond.beginAtomIdx

        for bond in new_mol.atoms[p_b_idx].bonds:  # should only ever be one bond for any pseudoatom
            if bond.beginAtomIdx == p_b_idx:
                b_idx = bond.endAtomIdx
            else:
                b_idx = bond.beginAtomIdx

        # now make the new bonds
        # a
        for i, bond in enumerate(new_mol.atoms[a_idx].bonds):
            if (bond.beginAtomIdx, bond.endAtomIdx) == (a_idx, p_a_idx):
                new_mol.atoms[a_idx].bonds[i].endAtomIdx = b_idx
                break
            elif (bond.beginAtomIdx, bond.endAtomIdx) == (p_a_idx, a_idx):
                new_mol.atoms[a_idx].bonds[i].beginAtomIdx = b_idx
                break
        # b
        for i, bond in enumerate(new_mol.atoms[b_idx].bonds):
            if (bond.beginAtomIdx, bond.endAtomIdx) == (b_idx, p_b_idx):
                new_mol.atoms[b_idx].bonds[i].endAtomIdx = a_idx
                break
            elif (bond.beginAtomIdx, bond.endAtomIdx) == (p_b_idx, b_idx):
                new_mol.atoms[b_idx].bonds[i].beginAtomIdx = a_idx
                break

        # now remove the pseudoatoms
        del new_mol.atoms[p_a_idx]
        del new_mol.atoms[p_b_idx]
        del new_pi[p_atom1]
        del new_pi[p_atom2]
        new_mol.pseudoIndex = new_pi
        new_mol.brics_legend = new_bl

        return new_mol

    def get_fc1_frags(self):
        """
        This function gets (and sets) all the FC1 fragments. This is another special case, since you can just combine
        all the pairs of FC0 fragments that match, and that is definitively all the FC1 fragments.
        :return:
        """
        fc1_frags = []
        for p_atom1, p_atom2 in self.edges:  # for every bond (now broken)
            # get the FC0 fragments
            frag1 = self.fragments[0][self.fc0_index[p_atom1]]
            frag2 = self.fragments[0][self.fc0_index[p_atom2]]
            # combine them
            new_mol = self.react(frag1=frag1, frag2=frag2, p_atom1=p_atom1, p_atom2=p_atom2)
            if new_mol:
                if new_mol not in self.frag_cache:
                    fc1_frags.append(new_mol)
                    self.frag_cache.append(new_mol)
        self.fragments[1] = fc1_frags

    def _copyAtom(self, atom):
        newBonds = [self.Bond(bond.bondType, bond.beginAtomIdx, bond.endAtomIdx) for bond in atom.bonds]
        return self.Atom(index=atom.index, atom_num=atom.atom_num, isotope=atom.isotope,
                         chiralTag=atom.chiralTag, explicitHs=atom.explicitHs,
                         bonds=newBonds, aromaticTag=atom.aromaticTag)

    def get_next_fc_level_frags(self, *, fc):
        """
        Given an FC value for which the fragments already exist, compute the next level up of complexity
        :param fc:
        :return: nothing, edit in place
        """
        new_frags = []
        for old_frag in self.fragments[fc]:
            p_atoms = deepcopy(list(old_frag.pseudoIndex.keys()))
            for p_atom_label in p_atoms:
                if p_atom_label % 2 == 0:
                    other_p_atom = p_atom_label+1
                else:
                    other_p_atom = p_atom_label - 1
                frag2 = self.fragments[0][self.fc0_index[other_p_atom]]
                new_mol = self.react(frag1=old_frag, frag2=frag2, p_atom1=p_atom_label, p_atom2=other_p_atom)
                # it is possible to have multiple paths create the same fragment
                # detect here and ignore (this keeps the search space much smaller)
                if new_mol not in self.frag_cache:
                    new_frags.append(new_mol)
                    self.frag_cache.append(new_mol)
        self.fragments[fc+1] = new_frags

    def get_all_fragments(self, max_fc=7):
        """
        Given a target FC value, calculate all the fragments up to and including that FC. Format them in a way that is
        ready to be written to the database.
        :param max_fc:
        :return: frag_list, heritage_list, atoms_list, pseudo_atoms_list (lists of dictionaries matching models)
        """
        self.get_fc0_frags()  # always do this first
        max_fc = min(len(self.edges), max_fc)
        if len(self.edges) > 1:  # if there is more than one brics bond
            self.get_fc1_frags()
        if len(self.edges) > 2:  # if there are more than two brics bonds
            for i in range(1, max_fc):
                self.get_next_fc_level_frags(fc=i)

        # these will store the dictionaries the database code expects
        frag_list = []
        heritage_list = []
        atoms_list = []
        pseudo_atoms_list = []

        # using this function to avoid copying code
        def insert_mol_info(mol, fc, curr_id, parent_id):

            seen_ps = set()
            seen_atom = set()
            num_pseudo_atoms = 0
            num_unique_pseudo_atoms = 0
            has_only_good_atoms = True

            # sanity checking to see if this is a fragment we want to use for making molecules
            for atom in mol.GetAtoms():
                symbol = atom.GetSymbol()
                if symbol not in allowed_atoms:
                    has_only_good_atoms = False
                    break
                atomic_num = atom.GetAtomicNum()
                ps_num = atom.GetIsotope()

                # storing info about each atom
                if symbol not in seen_atom and atomic_num != 0:
                    atoms_list.append({
                        'frag': curr_id,
                        'atom': symbol
                    })
                    seen_atom.add(symbol)
                if atomic_num == 0:
                    num_pseudo_atoms += 1
                    pseudo_atoms_list.append({
                        'frag': curr_id,
                        'pseudo_atom': ps_num
                    })
                    if ps_num not in seen_ps:
                        num_unique_pseudo_atoms += 1
                        seen_ps.add(ps_num)

            # yeah we'll take it
            if has_only_good_atoms:
                frag_list.append({
                    'id': curr_id,
                    'smile': Chem.MolToSmiles(mol, isomericSmiles=True),
                    'frag_coeff': fc,
                    'num_pseudo_atoms': num_pseudo_atoms,
                    'num_unique_pseudo_atoms': num_unique_pseudo_atoms
                })
                heritage_list.append({
                    'frag': curr_id,
                    'parent': parent_id
                })

        # insert parent first, since the fragments refer to the parent_id
        parent_id = ID_COUNTER.increment()
        curr_id = parent_id
        insert_mol_info(self.init_mol, len(self.edges), curr_id, parent_id)  # parents have a frag_id == parent_id
        curr_id = ID_COUNTER.increment()
        # insert all sub fragments
        for fc in range(max_fc):
            frag_group = self.fragments[fc]
            for frag in frag_group:
                insert_mol_info(frag.toMol(), fc, curr_id, parent_id)
                curr_id = ID_COUNTER.increment()
        return frag_list, heritage_list, atoms_list, pseudo_atoms_list

    def toMol(self):
        """
        Turn a RecomposerMol object back into an RDKIT Chem.Mol object, with correct stereochemistry
        and pseudoatoms if relevant
        :return: Chem.Mol
        """
        emol = Chem.RWMol(Chem.MolFromSmiles(""))  # create an empty editable mol
        atomsIndices = sorted(list(self.atoms.keys()))
        idxMapping = {}
        # add all the atoms back
        for i, idx in enumerate(atomsIndices):
            atom = self.atoms[idx]
            newAtom = Chem.Atom(atom.atom_num)
            if atom.atom_num == 0:
                newAtom.SetIsotope(self.brics_legend[atom.isotope])
            else:
                newAtom.SetIsotope(atom.isotope)
            newAtom.SetNoImplicit(True)
            newAtom.SetNumExplicitHs(atom.explicitHs)
            newAtom.SetChiralTag(atom.chiralTag)
            newAtom.SetIsAromatic(atom.aromaticTag)
            emol.AddAtom(newAtom)
            idxMapping[idx] = i
        # add all the bonds back
        for idx in atomsIndices:
            for bond in self.atoms[idx].bonds:
                begin = bond.beginAtomIdx
                end = bond.endAtomIdx
                if idx == begin:
                    emol.AddBond(idxMapping[begin], idxMapping[end], bond.bondType)
        return emol.GetMol()


def libgen(mol_list, output_name):
    """
    function to generate a database format library of fragments from a mol, list of mol objects, .smi, or .sdf file
    :param mol_list: list of molecules, a single molecule, or a filename of molecules to read
    :type mol_list: str|Chem.Mol|[Chem.Mol]
    :param output_name: name of the database to use?
    :type output_name: str
    :return:
    """
    # if a file not a list then read into list
    if isinstance(mol_list, str) and mol_list.endswith(".smi"):
        mol_list = Chem.SmilesMolSupplier(mol_list, delimiter="\t", titleLine=False)
    elif isinstance(mol_list, str) and mol_list.endswith(".sdf"):
        mol_list = Chem.SDMolSupplier(mol_list)
    elif type(mol_list) == Chem.Mol:
        mol_list = [mol_list]
    elif type(mol_list) == list:
        assert type(mol_list[0]) == Chem.Mol
    else:
        raise Exception("Did you provide a list of mol objects? Input type error.")

    fragment_dict_deque = deque()
    heritage_dict_deque =deque()
    atoms_dict_deque = deque()
    pseudoatoms_dict_deque = deque()
    logger.info("Fragmenting:")
    n = len(mol_list)
    i = 0
    t0 = time.time()
    for mol in mol_list:
        re_mol = RecomposerMol.fromMol(mol=mol)
        frag_list, heritage_list, atoms_list, pseudo_atoms_list = re_mol.get_all_fragments(7)
        fragment_dict_deque.extend(frag_list)
        heritage_dict_deque.extend(heritage_list)
        atoms_dict_deque.extend(atoms_list)
        pseudoatoms_dict_deque.extend(pseudo_atoms_list)
        logger.info("DONE: %d/%d %.f" % (
            i, n, 1000 * (time.time() - t0) / (i + 1)
        ))
        i += 1

    logger.info("Done")
    logger.info("Saving:")

    # create the database for the output
    db = SqliteExtDatabase(output_name, pragmas={
        'cache_size': -1024 * 64,  # 64MB page-cache.
        'journal_mode': 'wal',  # Use WAL-mode (you should always use this!).
        'foreign_keys': 0,
        'wal_autocheckpoint': 10,
    })

    db.connect()
    # get the models
    Fragment, Heritage, PseudoAtoms, Atoms = lib_read(db)

    Fragment.create_table(safe=True)
    Heritage.create_table(safe=True)
    PseudoAtoms.create_table(safe=True)
    Atoms.create_table(safe=True)
    with db.atomic():
        if len(fragment_dict_deque) > 0:
            for ents in chunked(fragment_dict_deque, 200):
                query = Fragment.replace_many(ents)
                query.execute()
            for ents in chunked(heritage_dict_deque, 200):
                query = Heritage.replace_many(ents)
                query.execute()
            for ents in chunked(pseudoatoms_dict_deque, 200):
                query = PseudoAtoms.replace_many(ents)
                query.execute()
            for ents in chunked(atoms_dict_deque, 200):
                query = Atoms.replace_many(ents)
                query.execute()
    db.close()
    clean(output_name)

    return 1


def get_pseudo_atoms(smile):
    """
    function to get BRICS pseudoatoms from a smile, and return the number of pseudoatoms, the num unique, and the atoms
    :param smile:
    :return:
    """
    num_pseudo_atoms = 0
    num_unique_pseudo_atoms = 0
    pseudo_atoms = []
    pattern = re.compile(r'\[([0-9]+)\*\]')
    try:
        pseudo_atoms = pattern.findall(smile)
        num_pseudo_atoms = len(pseudo_atoms)
        num_unique_pseudo_atoms = len(set(pseudo_atoms))
    except Exception:  # pylint: disable=broad-except
        pass

    return num_pseudo_atoms, num_unique_pseudo_atoms, pseudo_atoms


if __name__ == "__main__":
    # one of the only CLIs here. Just for convenience. Calls libgen on an input file. Now automatically cleaned
    parser = ArgumentParser()
    parser.add_argument('-i',
                        dest="input",
                        type=str,
                        help='.smi or .sdf file of molecules to be fragmented.',
                        required=True)
    parser.add_argument("-o",
                        dest="output",
                        type=str,
                        help="Database location for new fragment library.",
                        required=True)

    options = parser.parse_args()
    runtime = libgen(options.input, options.output)
    logger.info(runtime)



from setuptools import setup
from setuptools import find_packages

setup(name='deriver',
      description='Deriver: for all your molecule generation needs.',
      long_description='A software tool for the generation of novel chemical entities.',
      version='2.4.0',
      url='https://github.com/cyclica/deriver',
      license='All Rights Reserved Cyclica Inc.',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=False,
      install_requires=["selfies==1.0.1", "peewee>=3.9.3", "numpy>=1.16.3"]
      )



"""
Written by Jan H. Jensen 2018, copied 04/2020
"""
from rdkit import Chem
from rdkit.Chem import AllChem

import random

from rdkit import rdBase
rdBase.DisableLog('rdApp.error')


def cut(mol):
    if not mol.HasSubstructMatch(Chem.MolFromSmarts('[*]-;!@[*]')): 
        return None
    bis = random.choice(mol.GetSubstructMatches(Chem.MolFromSmarts('[*]-;!@[*]')))  # single bond not in ring
    bs = [mol.GetBondBetweenAtoms(bis[0],bis[1]).GetIdx()]
  
    fragments_mol = Chem.FragmentOnBonds(mol,bs,addDummies=True,dummyLabels=[(1, 1)])
  
    try:
        fragments = Chem.GetMolFrags(fragments_mol,asMols=True)
        return fragments
    except:
        return None


def cut_ring(mol):
    for i in range(10):
        if random.random() < 0.5:
            if not mol.HasSubstructMatch(Chem.MolFromSmarts('[R]@[R]@[R]@[R]')): 
                return None
            bis = random.choice(mol.GetSubstructMatches(Chem.MolFromSmarts('[R]@[R]@[R]@[R]')))
            bis = ((bis[0],bis[1]),(bis[2],bis[3]),)
        else:
            if not mol.HasSubstructMatch(Chem.MolFromSmarts('[R]@[R;!D2]@[R]')): 
                return None
            bis = random.choice(mol.GetSubstructMatches(Chem.MolFromSmarts('[R]@[R;!D2]@[R]')))
            bis = ((bis[0],bis[1]),(bis[1],bis[2]),)

        bs = [mol.GetBondBetweenAtoms(x,y).GetIdx() for x,y in bis]
    
        fragments_mol = Chem.FragmentOnBonds(mol,bs,addDummies=True,dummyLabels=[(1, 1),(1,1)])
    
        try:
            fragments = Chem.GetMolFrags(fragments_mol,asMols=True)
        except:
            return None
    
        if len(fragments) == 2:
            return fragments
      
    return None


def ring_OK(mol):
    if not mol.HasSubstructMatch(Chem.MolFromSmarts('[R]')):
        return True
    
    ring_allene = mol.HasSubstructMatch(Chem.MolFromSmarts('[R]=[R]=[R]'))
    
    cycle_list = mol.GetRingInfo().AtomRings() 
    max_cycle_length = max([ len(j) for j in cycle_list ])
    macro_cycle = max_cycle_length > 6
    
    double_bond_in_small_ring = mol.HasSubstructMatch(Chem.MolFromSmarts('[r3,r4]=[r3,r4]'))
    
    return not ring_allene and not macro_cycle and not double_bond_in_small_ring


def crossover_ring(parent_A,parent_B):
    ring_smarts = Chem.MolFromSmarts('[R]')
    if not parent_A.HasSubstructMatch(ring_smarts) and not parent_B.HasSubstructMatch(ring_smarts):
      return None
    
    rxn_smarts1 = ['[*:1]~[1*].[1*]~[*:2]>>[*:1]-[*:2]','[*:1]~[1*].[1*]~[*:2]>>[*:1]=[*:2]']
    rxn_smarts2 = ['([*:1]~[1*].[1*]~[*:2])>>[*:1]-[*:2]','([*:1]~[1*].[1*]~[*:2])>>[*:1]=[*:2]']
    for i in range(10):
        fragments_A = cut_ring(parent_A)
        fragments_B = cut_ring(parent_B)
        if fragments_A is None or fragments_B is None:
            return None
        
        new_mol_trial = []
        for rs in rxn_smarts1:
            rxn1 = AllChem.ReactionFromSmarts(rs)
            new_mol_trial = []
            for fa in fragments_A:
                for fb in fragments_B:
                    new_mol_trial.append(rxn1.RunReactants((fa,fb))[0]) 
    
        new_mols = []
        for rs in rxn_smarts2:
            rxn2 = AllChem.ReactionFromSmarts(rs)
            for m in new_mol_trial:
                m = m[0]
                new_mols += list(rxn2.RunReactants((m,)))
          
        new_mols2 = []
        for m in new_mols:
            m = m[0]
            if ring_OK(m):
                new_mols2.append(m)
          
        if len(new_mols2) > 0:
            return random.choice(new_mols2)
      
    return None


def crossover_non_ring(parent_A,parent_B):
    for i in range(10):
        fragments_A = cut(parent_A)
        fragments_B = cut(parent_B)
        if fragments_A is None or fragments_B is None:
            return None
        rxn = AllChem.ReactionFromSmarts('[*:1]-[1*].[1*]-[*:2]>>[*:1]-[*:2]')
        new_mol_trial = []
        for fa in fragments_A:
            for fb in fragments_B:
                new_mol_trial.append(rxn.RunReactants((fa,fb))[0]) 
                                     
        new_mols = []
        for mol in new_mol_trial:
            mol = mol[0]
            new_mols.append(mol)
        
        if len(new_mols) > 0:
          return random.choice(new_mols)
        
    return None


def crossover(parent_A,parent_B):
    parent_smiles = [Chem.MolToSmiles(parent_A), Chem.MolToSmiles(parent_B)]
    try:
        Chem.Kekulize(parent_A, clearAromaticFlags=True)
        Chem.Kekulize(parent_B, clearAromaticFlags=True)
    except:
        pass
    for i in range(10):
        if random.random() <= 0.5:
            new_mol = crossover_non_ring(parent_A,parent_B)
            if new_mol is not None:
                new_smiles = Chem.MolToSmiles(new_mol)
            if new_mol is not None and new_smiles not in parent_smiles:
                return new_mol
    else:
        new_mol = crossover_ring(parent_A,parent_B)
        if new_mol is not None:
            new_smiles = Chem.MolToSmiles(new_mol)
        if new_mol is not None and new_smiles not in parent_smiles:
            return new_mol

    return None



from peewee import IntegerField, CharField, Model, ForeignKeyField, CompositeKey


def lib_read(lib_database):
    """
    format/models for the fragment databases
    :param lib_database:
    :return:
    """

    # a model for the fragments themselves
    class Fragment(Model):
        id = IntegerField(primary_key=True)
        frag_coeff = IntegerField()
        smile = CharField(unique=True)
        num_pseudo_atoms = IntegerField(default=0)
        num_unique_pseudo_atoms = IntegerField(default=0)

        class Meta:
            table_name = 'fragments'
            database = lib_database

    # a model for an entry in the heritage table, which contains information
    # about which fragments came from which molecules
    class Heritage(Model):
        frag = ForeignKeyField(
            Fragment, on_delete="CASCADE", on_update="CASCADE")
        parent = ForeignKeyField(Fragment, null=True)

        class Meta:
            table_name = 'heritage'
            database = lib_database
            primary_key = CompositeKey('frag', 'parent')

    # a model for pseudoatoms which enables searching for specific fragments by pseudoatom type
    class PseudoAtoms(Model):
        frag = ForeignKeyField(
            Fragment, on_delete="CASCADE", on_update="CASCADE")
        pseudo_atom = IntegerField(null=True)

        class Meta:
            table_name = 'pseudoatoms'
            database = lib_database
            primary_key = CompositeKey('frag', 'pseudo_atom')

    # a model for atoms, much like the heritage link table, which enables searching for specific fragments by atom
    class Atoms(Model):
        frag = ForeignKeyField(
            Fragment, on_delete="CASCADE", on_update="CASCADE")
        atom = CharField(null=True)

        class Meta:
            table_name = 'atoms'
            database = lib_database
            primary_key = CompositeKey('frag', 'atom')

    return Fragment, Heritage, PseudoAtoms, Atoms



# coding=utf-8
#
# Wrappers for query execution, with different levels of caching
#
# script by Andreas Windemuth

from typing import Callable, Any
import numpy as np
from .util import Vec, percent
from .fragment_queries import FragArray, MateQueryParams


class CacheL2(object):
    """Level 2 caching"""

    def __init__(self, fragments: FragArray, query):
        self.fragments = fragments
        self.query = query
        self.cache = Cache(self.l1_calc)

    def l1_calc(self, params):
        (ix,) = np.where(self.query.clause(self.fragments, params))
        pseudo = self.fragments.pseudo[ix]
        cache2 = Cache(self.l2_calc)
        return ix, pseudo, cache2

    def l2_calc(self, data):
        (pseudo, mask) = data
        (iy,) = np.where((pseudo & mask) != 0)
        return iy

    def select(self, params: MateQueryParams) -> Vec[int]:
        key1 = self.query.key(params)
        ix, pseudo, cache2 = self.cache.get(key1, params)
        key2 = params.mask
        iy = cache2.get(key2, (pseudo, params.mask))
        return ix[iy]

    def info(self):
        nh = 0
        nm = 0
        for _, _, c in self.cache.store.values():
            nh += c.n_hits
            nm += c.n_misses
        if nh + nm > 0:
            s = " %.2f%%" % (100.0 * nm / (nh + nm))
        else:
            s = ""
        return self.cache.info() + s


class Cache(object):
    """A basic general cache object. May put this into base/util later"""

    def __init__(self, calc: Callable[[Any], Any] = None):
        """
        Initializa a cache.
        :param calc: Function that will compute y from x
        """
        self.calc = calc
        self.store = {}
        self.n_hits = 0
        self.n_misses = 0
        self.size = 0

    def get(self, key, x=None):
        """Retrieve from cache if found, else compute and store"""
        if key in self.store:
            self.n_hits += 1
            y = self.store[key]
        else:
            self.n_misses += 1
            if self.calc is None:
                y = None
            else:
                y = self.calc(x)
                self.store[key] = y
        return y

    def set(self, key, y):
        self.store[key] = y

    def info(self):
        return percent(self.n_misses, self.n_misses + self.n_hits)



import logging as logger
import uuid
from .child_filter import get_filter_values, apply_filter
from .config import drug_like_params
from rdkit import Chem
from .selfies_methods import selfies_substitution, selfies_deletion, selfies_insertion, random_selfies_generator, \
    selfies_scanner
from typing import List, Set
from collections import defaultdict
from .fragment import libgen
from peewee import SqliteDatabase
from .lib_read import lib_read
import re
from .fragment_index import frag_index
from .mate import mate
import os
import random
from .jensen_crossover import crossover as crossover_gb
from .jensen_mutate import mutate as mutate_gb
from .jensen_selfies_crossover import crossover as selfies_crossover_gb
from .jensen_selfies_mutate import mutate as selfies_mutate_gb


class Deriver(object):

    def __init__(self):

        logger.info("Initializing a new Deriver object!")
        self.data = self._Data()

    class _Data(object):

        """
        This object is meant to be a home for all the data and parameters used by deriver.
        It should be fine to read things from here, but do not set anything directly. Instead,
        use the methods that begin with "set_" in order to change these values. This
        prevents incompatible changes from being made silently.
        """

        def __init__(self):
            self.seed_smiles = None
            self.seed_mols = None
            self.filter_params = drug_like_params
            self.filter = False
            self.child_db = None
            self.all_good_selfies_children = None
            self.all_good_scanner_children = None
            self.all_good_selfies_gb_children = None
            self.all_good_smiles_gb_children = None
            self.filter_molecules = None
            self.must_have_patterns = None
            self.must_not_have_patterns = None
            self.heritage = defaultdict(list)
            # BRICS specific
            self.seed_frags = None  # these are the fragments of the seed molecules
            self.fragment_source_db = None  # this is the location of the fragment DB
            self.seed_frag_db = None  # the is the DB where the seed_frags are stored and info about them
            self.all_good_brics_children = None  # this is where the good (filtered) BRICS children are saved

    def set_seeds(self, seeds: list):

        """
        set the seeds that are used to generate new molecules
        :param seeds:
        :return:
        """

        if isinstance(seeds[0], Chem.rdchem.Mol):
            self.data.seed_smiles = []
            self.data.seed_mols = []
            for seed in seeds:
                smile = Chem.MolToSmiles(seed, isomericSmiles=True)
                self.data.seed_smiles.append(smile)
                self.data.seed_mols.append(seed)
        elif isinstance(seeds[0], str):
            self.data.seed_smiles = []
            self.data.seed_mols = []
            for smile in seeds:
                seed = Chem.MolFromSmiles(smile, sanitize=True)
                iso_smile = Chem.MolToSmiles(seed, isomericSmiles=True)
                self.data.seed_smiles.append(iso_smile)
                self.data.seed_mols.append(seed)
        else:
            logger.error("Seeds must be provided as an iterable of Mol objects, or SMILES strings.")
            return 0

        return 1

    def set_must_have_patterns(self, must_have_patterns: List[str]):
        if isinstance(must_have_patterns, list):
            assert isinstance(must_have_patterns[0], str)
        elif must_have_patterns is None:
            pass
        else:
            raise TypeError("must_have_patterns must be None or a list of SMARTS strings")
        self.data.must_have_patterns = must_have_patterns

    def set_must_not_have_patterns(self, must_not_have_patterns: List[str]):
        if isinstance(must_not_have_patterns, list):
            assert isinstance(must_not_have_patterns[0], str)
        elif must_not_have_patterns is None:
            pass
        else:
            raise TypeError("must_have_patterns must be None or a list of SMARTS strings")
        self.data.must_not_have_patterns = must_not_have_patterns

    def set_filter_molecules(self, filter_molecules: Set[str]):
        assert isinstance(filter_molecules, set)
        assert isinstance(iter(filter_molecules).__next__(), str)
        self.data.filter_molecules = filter_molecules
        return 1

    def enable_and_expand_filter(self, seeds: list = None):

        """
        Turn on the filtering of produced molecules, and expand the ranges to include the provided
        seed molecules, or the stored seed molecules.
        :param seeds:
        :return:
        """

        if seeds is None:
            if self.data.seed_smiles is None:
                logger.info("Turning on filter using default filter values:")
                logger.info(self.data.filter_params)
                self.data.filter = True
                return 1
            else:
                logger.info(f"Expanding filter ranges using current seed set: {self.data.seed_smiles}")
                seeds = self.data.seed_mols
        for seed_mol in seeds:
            parent_values = get_filter_values(seed_mol)
            self._update_filter_params(parent_values)
        logger.info(f"Done! Turning on filter using new filter values:")
        logger.info(self.data.filter_params)
        self.data.filter = True
        return 1

    def manual_filter_set(self, var: str, val1, val2=None):

        """
        Use this function to change the filter values safely.
        :param var:
        :param val1:
        :param val2:
        :return:
        """

        if var not in self.data.filter_params:
            logger.error(f"{var} is not a valid filter parameter, try again.")
            raise Exception

        logger.info(f"Updating {var} from {self.data.filter_params[var]} to {val1}, {val2}")

        if isinstance(self.data.filter_params[var], tuple):

            if val2 is None:
                logger.error(f"{var} requires TWO values, lower and upper bounds, to be set.")
                raise Exception

            elif val2 <= val1:
                logger.error(f"{var} requires TWO values, lower and upper bounds, to be set. "
                             f"The second must be LARGER than the first.")
                raise Exception

            else:
                # actually set the values!
                if isinstance(self.data.filter_params[var][0], int):
                    val1 = int(val1)
                    val2 = int(val2)
                self.data.filter_params[var] = (val1, val2)

        else:
            # here set these values
            if isinstance(self.data.filter_params[var], int):
                val1 = int(val1)
            self.data.filter_params[var] = val1

        logger.info(f"Done! Current filter parameters are: {self.data.filter_params}")

        return 1

    def _update_filter_params(self, parent_values):

        filter_params = {
            # ranges
            "MW": (min(parent_values["MW"], self.data.filter_params["MW"][0]),
                   max(parent_values["MW"], self.data.filter_params["MW"][1])),
            "num_carbons": (min(parent_values["num_carbons"], self.data.filter_params["num_carbons"][0]),
                            max(parent_values["num_carbons"], self.data.filter_params["num_carbons"][1])),
            "num_hetero_atoms": (min(parent_values["num_hetero_atoms"],
                                     self.data.filter_params["num_hetero_atoms"][0]),
                                 max(parent_values["num_hetero_atoms"],
                                     self.data.filter_params["num_hetero_atoms"][1])),
            "hc_ratio": (min(parent_values["hc_ratio"], self.data.filter_params["hc_ratio"][0]),
                         max(parent_values["hc_ratio"], self.data.filter_params["hc_ratio"][1])),
            "charge": (min(parent_values["charge"], self.data.filter_params["charge"][0]),
                       max(parent_values["charge"], self.data.filter_params["charge"][1])),
            "logP": (min(parent_values["logP"], self.data.filter_params["logP"][0]),
                     max(parent_values["logP"], self.data.filter_params["logP"][1])),

            # upper limits
            "HBA": max(parent_values["HBA"], self.data.filter_params["HBA"]),
            "HBD": max(parent_values["HBD"], self.data.filter_params["HBD"]),
            "tPSA": max(parent_values["tPSA"], self.data.filter_params["tPSA"]),
            "rot_bonds": max(parent_values["rot_bonds"], self.data.filter_params["rot_bonds"]),
            "rigid_bonds": max(parent_values["rigid_bonds"], self.data.filter_params["rigid_bonds"]),
            "num_rings": max(parent_values["num_rings"], self.data.filter_params["num_rings"]),
            "max_ring_size": max(parent_values["max_ring_size"], self.data.filter_params["max_ring_size"]),
            "num_charges": max(parent_values["num_charges"], self.data.filter_params["num_charges"]),
            "num_chiral_centers": max(parent_values["num_chiral_centers"],
                                      self.data.filter_params["num_chiral_centers"])

        }

        self.data.filter_params = filter_params

    def derive_selfies(self, n_children: int = 100, mut_rate: float = 0.03, mut_min: int = 1, mut_max: int = 2):

        good_children = []
        all_filtered_children = {}
        self.data.all_good_selfies_children = []
        n_seeds = len(self.data.seed_smiles)
        if n_children < n_seeds:
            n_children_per_seed = 1
        else:
            n_children_per_seed = round(n_children / n_seeds) + 1

        logger.info(f"Mutating SELFIES to create {n_children_per_seed} children per seed.")

        if self.data.filter:
            filter_params = self.data.filter_params
        else:
            logger.warning("Warning: No filter has been set, so all child molecules will be labeled"
                           " as 'good' regardless of quality. Please call Deriver.set_filter() first"
                           " in order to use a filter for drug-likeness.")
            filter_params = None

        for seed in self.data.seed_smiles:
            children = selfies_substitution(parent_smiles=seed,
                                            n_children=round(n_children_per_seed * 0.7),  # three internal methods
                                            mut_rate=mut_rate,
                                            mut_min=mut_min,
                                            mut_max=mut_max)
            self.data.heritage[seed] += children
            child_mols = [Chem.MolFromSmiles(child, sanitize=True) for child in children]

            children = selfies_insertion(parent_smiles=seed,
                                         n_children=round(n_children_per_seed * 0.15),  # three internal methods
                                         mut_rate=mut_rate,
                                         mut_min=mut_min,
                                         mut_max=mut_max)
            self.data.heritage[seed] += children
            child_mols += [Chem.MolFromSmiles(child, sanitize=True) for child in children]

            children = selfies_deletion(parent_smiles=seed,
                                        n_children=round(n_children_per_seed * 0.15),  # three internal methods
                                        mut_rate=mut_rate,
                                        mut_min=mut_min,
                                        mut_max=mut_max)
            self.data.heritage[seed] += children
            child_mols += [Chem.MolFromSmiles(child, sanitize=True) for child in children]

            # filter children
            filtered_children = apply_filter(filter_params,
                                             child_mols,
                                             self.data.must_have_patterns,
                                             self.data.must_not_have_patterns
                                             )
            all_filtered_children.update(filtered_children)

            for child in filtered_children:
                if filtered_children[child]["is_good"]:
                    # check the cache
                    if self.data.filter_molecules:
                        if child not in self.data.filter_molecules:
                            good_children.append(child)
                        else:
                            logger.debug(f"skipping previously seen molecule: {child}")
                    else:
                        good_children.append(child)

        logger.info(f"Generated {len(good_children)} 'good' children.")
        self.data.all_good_selfies_children = good_children
        return good_children, all_filtered_children

    def random_selfies(self, *, n_symbols: int = 100, n_molecules: int = 100):

        good_children = []
        rand_selfies_gen = random_selfies_generator(n_symbols=n_symbols)
        self.data.all_good_selfies_children = []

        if self.data.filter:
            filter_params = self.data.filter_params
        else:
            logger.warning("Warning: No filter has been set, so all child molecules will be labeled"
                           " as 'good' regardless of quality. Please call Deriver.set_filter() first"
                           " in order to use a filter for drug-likeness.")
            filter_params = None

        while len(good_children) < n_molecules:
            child_mols = [Chem.MolFromSmiles(next(rand_selfies_gen)) for i in range(n_molecules - len(good_children))]
            # filter children
            filtered_children = apply_filter(filter_params,
                                             child_mols,
                                             self.data.must_have_patterns,
                                             self.data.must_not_have_patterns
                                             )

            for child in filtered_children:
                if filtered_children[child]["is_good"]:
                    # check the cache
                    if self.data.filter_molecules:
                        if child not in self.data.filter_molecules:
                            good_children.append(child)
                        else:
                            logger.debug(f"skipping previously seen molecule: {child}")
                    else:
                        good_children.append(child)

        self.data.all_good_selfies_children = good_children
        return good_children

    def scan_selfies(self):

        """
        Return all possible single substitution children for all the seeds.
        """
        if self.data.filter:
            filter_params = self.data.filter_params
        else:
            logger.warning("Warning: No filter has been set, so all child molecules will be labeled"
                           " as 'good' regardless of quality. Please call Deriver.set_filter() first"
                           " in order to use a filter for drug-likeness.")
            filter_params = None
        good_children = []
        self.data.all_good_scanner_children = []
        all_filtered_children = {}

        for seed in self.data.seed_smiles:
            children = selfies_scanner(parent_smiles=seed)
            self.data.heritage[seed] += children

            filtered_children = apply_filter(filter_params,
                                             [Chem.MolFromSmiles(child) for child in children],
                                             self.data.must_have_patterns,
                                             self.data.must_not_have_patterns
                                             )
            all_filtered_children.update(filtered_children)

            for child in filtered_children:
                if filtered_children[child]["is_good"]:
                    # check the cache
                    if self.data.filter_molecules:
                        if child not in self.data.filter_molecules:
                            good_children.append(child)
                        else:
                            logger.debug(f"skipping previously seen molecule: {child}")
                    else:
                        good_children.append(child)

        self.data.all_good_scanner_children = good_children
        return good_children, all_filtered_children

    def derive_gb(self, n_children: int = 100, representation='selfies'):

        assert len(self.data.seed_smiles) > 0
        children = []
        good_children = []
        if representation == 'selfies':
            self.data.all_good_selfies_gb_children = []
            crossover_fn = selfies_crossover_gb
            mutation_fn = selfies_mutate_gb
        elif representation =='smiles':
            self.data.all_good_smiles_gb_children = []
            crossover_fn = crossover_gb
            mutation_fn = mutate_gb
        else:
            raise ValueError('Must specify derivation kind as one of "smiles" or "selfies"')

        if self.data.filter:
            filter_params = self.data.filter_params
        else:
            logger.warning("Warning: No filter has been set, so all child molecules will be labeled"
                           " as 'good' regardless of quality. Please call Deriver.set_filter() first"
                           " in order to use a filter for drug-likeness.")
            filter_params = None

        parent_a_smiles, parent_b_smiles = (None, None)
        if len(self.data.seed_smiles) > 1:
            do_crossover = True
            new_child = None
        else:
            do_crossover = False
            new_child = self.data.seed_mols[0]

        for _ in range(n_children):
            if do_crossover:
                parent_a, parent_b = random.sample(self.data.seed_mols, 2)
                new_child = crossover_fn(parent_a, parent_b)
            if new_child is not None:
                mutated_child = mutation_fn(new_child)
                if mutated_child is None:
                    continue
            else:
                continue
            children.append(mutated_child)

        filtered_children = apply_filter(filter_params,
                                         children,
                                         self.data.must_have_patterns,
                                         self.data.must_not_have_patterns
                                         )
        for child in filtered_children:
            if filtered_children[child]["is_good"]:
                # check the cache
                if self.data.filter_molecules:
                    if child not in self.data.filter_molecules:
                        good_children.append(child)
                    else:
                        logger.debug(f"skipping previously seen molecule: {child}")
                else:
                    good_children.append(child)

        if representation == 'smiles':
            self.data.all_good_smiles_gb_children = good_children
        else:
            self.data.all_good_selfies_gb_children = good_children
        logger.info(f"Generated {len(good_children)} 'good' children.")
        return good_children, filtered_children

    def set_fragment_source_db(self, frag_db):

        """
        set the location for the fragment database that is used to mate molecules
        :param frag_db:
        :return:
        """

        self.data.fragment_source_db = frag_db
        return 1

    def _process_seeds_for_brics(self):
        """
        This function parses the seed molecules and gets the BRICS fragments they make, then cleans them
        :return:
        """
        logger.info("Processing seeds to create scaffold fragments:")

        self.data.seed_frag_db = f"seed_frags_{uuid.uuid4()}.db"
        self.data.seed_frags = []

        # Databases are used in lieu of alternatives (like dataframes) in order to operate on larger datasets
        # where memory may be a bottleneck, and to ensure that only one source of truth exists for performing
        # this calculation.
        libgen(self.data.seed_mols, self.data.seed_frag_db)  # libgen is the basic command that generates frag libraries
        seed_frag_db = SqliteDatabase(self.data.seed_frag_db)  # load the db we just made
        Fragment, Heritage, _, _ = lib_read(seed_frag_db)  # we only care about fragment and heritage at this point
        seed_frag_db.connect()

        # get all the fragments from the user molecule
        user_frags = \
            (Fragment.select()
             .join(Heritage, on=Heritage.frag)
             .where(Heritage.parent != Heritage.frag)  # only get fragments, not intact molecules
             .order_by(Fragment.frag_coeff.desc()))  # largest and most complex frags first

        # for every fragment from the user provided parent mol
        for user_frag in user_frags:

            # we want to ignore really small fragments, by counting atom symbols
            smaller_smile = re.sub(r"\[[0-9]+\*\]", "", user_frag.smile)  # ignore pseudoatoms
            smaller_smile = re.sub(r"[0-9]", "", smaller_smile)  # ignore numbers in general
            smaller_smile = re.sub(r"[\(\)=#@\-\]\[]+", "", smaller_smile)  # ignore a bunch of other symbols
            if len(smaller_smile) < 4:  # if there are less than four atoms
                logger.warning(f"Skipping user_frag {user_frag.smile} due to size.")
                continue

            # using this fragment and the whole parent molecule, estimate the "missing" FC and size
            try:
                parent = Heritage.get(Heritage.frag_id == user_frag.id).parent
            # todo: actual exception is deriver.lib_read.FragmentDoesNotExist, check if we can except just this case
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(f'Encountered exception {e}')
                logger.warning('If this exception describes a missing parent in the Heritage table, this bug'
                               'is known and is being handled as intended.')
                continue

            # if parent is not None:
            missing_piece_fc = (parent.frag_coeff - user_frag.frag_coeff) - 1.0  # -1.0 because two pieces combine
            missing_piece_len = len(parent.smile) - len(user_frag.smile)  # approximation
            # else:
            #       missing_piece_fc = 3.0  # approximation
            #       missing_piece_len = 40  # approximation

            # this is what we are going to keep
            seed_frag = (user_frag.smile,
                         user_frag.num_pseudo_atoms,
                         missing_piece_fc,
                         missing_piece_len,
                         parent.smile)

            self.data.seed_frags.append(seed_frag)

        seed_frag_db.close()
        os.remove(self.data.seed_frag_db)  # we don't really care to keep the seed fragment database
        logger.info(f"Done! There are {len(self.data.seed_frags)} seed fragments ready to be mated.")
        # it is faster to keep these in memory rather than using the database

    def derive_brics(self, n_children: int = 100, permissivity: float = 1.0):
        """

        :param n_children: How many children do you want, in total. This is an approximation, not exact.
        :param permissivity: How unlike the parent molecules is the child allowed to be, higher is generally larger
        :return: (all_good_children [a list of smiles], all_filtered_children [a dict of values about the molecules])
        """
        # process the seeds
        self._process_seeds_for_brics()

        # get the "maximum number of children" per fragment
        n_seed_frags = len(self.data.seed_frags)

        if self.data.filter:
            filter_params = self.data.filter_params
        else:
            logger.warning("Warning: No filter has been set, so all child molecules will be labeled"
                           " as 'good' regardless of quality. Please call Deriver.set_filter() first"
                           " in order to use a filter for drug-likeness.")
            filter_params = None

        if n_children < n_seed_frags:
            children_per_seed_frag = 1
        else:
            children_per_seed_frag = round(n_children / n_seed_frags) + 1

        logger.info(f"Creating/reading a fragment index for {self.data.fragment_source_db}")
        # generate the frag index once, first, so it doesn't get generated in each pool process
        # this index serves to dramatically speed up queries
        frag_index(self.data.fragment_source_db)

        # again this is more of a guideline
        logger.info(f"Mating to create {children_per_seed_frag} children per seed frag.")

        all_filtered_children = dict()  # the filter returns a dictionary of calculated pk values and filter status

        for seed_frag_smile, seed_frag_num_pa, missing_p_fc, missing_p_len, parent_smile in self.data.seed_frags:
            try:
                res = mate(self.data.fragment_source_db,
                           seed_frag_smile,
                           seed_frag_num_pa,
                           missing_p_fc,
                           missing_p_len,
                           permissivity,
                           children_per_seed_frag,
                           filter_params,
                           self.data.must_have_patterns,
                           self.data.must_not_have_patterns)

                _, filter_values = res  # we only care about the filter dict, since it has everything
                all_filtered_children.update(filter_values)  # update our master dict
                self.data.heritage[parent_smile] += list(filter_values.keys())  # this keeps track of heritage
            except IndexError as e:
                # This bug has never really been explored that much.
                logger.warning(f"Error when trying to mate a molecule, ignoring this molecule. Error: {e}")

        all_good_children = []
        self.data.all_good_brics_children = []  # every time you call `derive_brics` it deletes any old results
        for child in all_filtered_children:
            if all_filtered_children[child]["is_good"]:
                # check the cache of previously seen molecules (which we want to avoid reproducing)
                if self.data.filter_molecules:
                    if child not in self.data.filter_molecules:
                        all_good_children.append(child)
                    else:
                        logger.debug(f"skipping previously seen molecule: {child}")
                else:
                    # there is no provided list of molecules to skip
                    all_good_children.append(child)

        logger.info(f"Generated {len(self.data.all_good_brics_children)} 'good' children.")
        self.data.all_good_brics_children = all_good_children
        return all_good_children, all_filtered_children



from typing import TypeVar, Union, Sequence
import numpy as np
from rdkit import Chem

# Vector and matrix types
T = TypeVar('T')
Mat = Union[np.ndarray, Union[T]]
Vec = Union[np.ndarray, Union[T]]


def percent(k: int, n: int) -> str:
    """Format a fraction with percentage"""
    if n == 0:
        pc = ""
    else:
        pc = " (%.1f%%)" % (100.0 * k / n)
    return "%d/%d" % (k, n) + pc


# Decorators
class lazy(object):
    """
    meant to be used for lazy evaluation of an object attribute.
    property should represent non-mutable data, as it replaces itself.
    Usage:
        class Foo:
            @lazy
            def longish_computation(self):
                ... longish computation
                return result

        foo = Foo()
        print("Result: "+foo.longish_computation)  # This will take some time
        print("Result: "+foo.longish_computation)  # This won't
    """

    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        setattr(obj, self.func_name, value)
        return value
